'use strict';

const config = require('./../../config');
const logger = require('./../../utils/logger');
const errors = require('./../../utils/dz-errors-api');
const dbConstants = require('./../../constants/db-constants');
const query = require('./../../utils/query-creator-api');
let async = require('async');
let _ = require('underscore');
let moment = require('moment');
const LD = require('lodash');
var days = ['Sun', 'Mon', 'Tue', 'Wed', 'Thu', 'Fri', 'Sat'];
let monthName = ["Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"];
const {
    forEach
} = require('p-iteration');
const timeZone = require('moment-timezone');
const notificationHandler = require('./../../model_handlers/api/notification-handler');

const {
    Configuration,
    PlaidApi,
    PlaidEnvironments,
    ProcessorStripeBankAccountTokenCreateRequest
} = require('plaid');

const configurationLive = new Configuration({
    basePath: PlaidEnvironments.production,
    baseOptions: {
        headers: {
            'PLAID-CLIENT-ID': config.plaid.client_id,
            'PLAID-SECRET': config.plaid.secret,
        },
    },
});

const configurationTest = new Configuration({
    basePath: PlaidEnvironments.sandbox,
    baseOptions: {
        headers: {
            'PLAID-CLIENT-ID': config.plaid.client_id,
            'PLAID-SECRET': config.plaid.test_secret,
        },
    },
});




/*
    Name : reservation list
    Purpose : reservation list
    Original Author : Gaurav Patel
    Created At : 06th APr 2021
*/
const reservationList = async (requestParam) => {
    return new Promise(async (resolve, reject) => {
        try {
            let response = await query.selectWithAndOne(dbConstants.dbSchema.customers, {
                customer_id: requestParam.customer_id
            }, {
                _id: 0,
                customer_id: 1
            });
            if (!response) {
                reject(errors.userNotFound(true, requestParam.code));
                return;
            }
            let matchColumn = {
                'owner_id': requestParam.customer_id
            }
            let today = timeZone(new Date()).tz(requestParam.time_zone).format('YYYY-MM-DD')
            if (requestParam.type == 'current') {
                matchColumn.status = { $in: ['parked'] }
            }
            if (requestParam.type == 'upcoming') {
                matchColumn.status = { $in: ['reserved'] }
            }
            if (requestParam.type == 'past') {
                matchColumn['$or'] = [{
                    status: 'completed'
                }, {
                    status: 'cancelled'
                }]
            };
            let joinArr = [{
                $lookup: {
                    from: 'customers',
                    localField: 'customer_id',
                    foreignField: 'customer_id',
                    as: 'seekerDetails'
                }
            }, {
                $unwind: "$seekerDetails"
            }, {
                $lookup: {
                    from: 'properties',
                    localField: 'property_id',
                    foreignField: 'property_id',
                    as: 'propDetails'
                }
            }, {
                $unwind: "$propDetails"
            }, {
                $lookup: {
                    from: 'parking_spot_types',
                    localField: 'propDetails.parking_spot_type_id',
                    foreignField: 'parking_spot_type_id',
                    as: 'spotTypesDetails'
                }
            }, {
                $unwind: "$spotTypesDetails"
            }, {
                $lookup: {
                    from: 'parking_spot_coverages',
                    localField: 'propDetails.parking_spot_coverage_id',
                    foreignField: 'parking_spot_coverage_id',
                    as: 'spotCoverageDetails'
                }
            }, {
                $unwind: "$spotCoverageDetails"
            }, {
                $lookup: {
                    from: 'vehicles',
                    localField: 'vehicle_id',
                    foreignField: 'vehicle_id',
                    as: 'vehDetails'
                }
            }, {
                $unwind: "$vehDetails"
            }, {
                $lookup: {
                    from: 'property_ratings',
                    localField: 'property_id',
                    foreignField: 'property_id',
                    as: 'ratingDetails'
                }
            }, {
                $lookup: {
                    from: 'chats',
                    localField: 'reservation_id',
                    foreignField: 'room',
                    as: 'chatDetails'
                }
            }, {
                "$unwind": {
                    "path": "$chatDetails",
                    "preserveNullAndEmptyArrays": true
                }
            }, {
                $match: matchColumn
            }, {
                $sort: {
                    created_at: -1
                }
            }, {
                $project: {
                    _id: 0,
                    reservation_id: 1,
                    property_id: 1,
                    status: 1,
                    from_date: 1,
                    to_date: 1,
                    parked_date: 1,
                    price: 1,
                    seeker_id: "$customer_id",
                    customer_id: "$propDetails.customer_id",
                    parking_spot_photos: "$propDetails.parking_spot_photos",
                    pricing: "$propDetails.pricing",
                    address: "$propDetails.address",
                    latitude: "$propDetails.latitude",
                    longitude: "$propDetails.longitude",
                    parking_size: "$propDetails.parking_size",
                    vehicle_photo: "$vehDetails.photo",
                    display_date: "",
                    parking_spot_type_title: "$spotTypesDetails.title." + requestParam.code,
                    parking_spot_coverage_title: "$spotCoverageDetails.title." + requestParam.code,
                    ratings: "$ratingDetails",
                    is_chat: "$chatDetails",
                    seeker_details: "$seekerDetails",
                }
            }];
            let reservation = await query.joinWithAnd(dbConstants.dbSchema.reservations, joinArr);
            reservation = JSON.parse(JSON.stringify(reservation))
            // let newArr = [];
            // if (requestParam.type == 'current') {
            //     _.each(reservation, (elem) => {
            //         let from_date = timeZone(new Date(elem.from_date)).tz(requestParam.time_zone).format('YYYY-MM-DD')
            //         if (today == from_date) {
            //             newArr.push(elem)
            //         }
            //     })
            //     reservation = newArr
            // }
            // if (requestParam.type == 'upcoming') {
            //     _.each(reservation, (elem) => {
            //         let from_date = timeZone(new Date(elem.from_date)).tz(requestParam.time_zone).format('YYYY-MM-DD')
            //         if (today != from_date) {
            //             newArr.push(elem)
            //         }
            //     })
            //     reservation = newArr
            // }
            await forEach(reservation, async (elem) => {
                elem.seeker_details = {
                    customer_id: elem.seeker_details.customer_id,
                    name: elem.seeker_details.name,
                    surname: elem.seeker_details.surname,
                    email: elem.seeker_details.email,
                    mobile_country_code: elem.seeker_details.mobile_country_code,
                    mobile: elem.seeker_details.mobile,
                    profile_picture: elem.seeker_details.profile_picture != '' ? config.aws.prefix + config.aws.s3.customerBucket + '/' + elem.seeker_details.profile_picture : '',
                }
                elem.vehicle_photo = elem.vehicle_photo != '' ? config.aws.prefix + config.aws.s3.vehicleBucket + '/' + elem.vehicle_photo : ''
                let arr = []
                _.each(elem.parking_spot_photos, (rec) => {
                    if (rec != '') {
                        arr.push(config.aws.prefix + config.aws.s3.propertyBucket + '/' + rec)
                    }
                });
                elem.parking_spot_photos = arr

                let rating = 0;
                if (elem.ratings.length > 0) {
                    rating = LD.sumBy(elem.ratings, 'rating')
                    rating = parseFloat(rating) / elem.ratings.length;
                }
                elem.rating = parseFloat(rating).toFixed(1);
                elem.out_of_rating = elem.ratings.length;
                delete elem.ratings

                if (requestParam.type != 'past') {
                    if (elem.status == 'reserved') {
                        elem.display_date = timeZone(new Date(elem.from_date)).tz(requestParam.time_zone).format('lll')
                    }
                    if (elem.status == 'parked') {
                        let todayDate = timeZone(new Date()).tz(config.time_zone);
                        let to_date = timeZone(new Date(elem.to_date)).tz(config.time_zone);
                        let duration = moment.duration(to_date.diff(todayDate));
                        const hours = parseInt(duration.asHours());
                        const minutes = parseInt(duration.asMinutes()) - hours * 60;
                        if (hours > 0) {
                            elem.display_date = hours + "hrs " + minutes + "mins"
                        } else {
                            elem.display_date = minutes + "mins"
                        }
                    }
                }
                elem.is_rated = false;
                console.log({
                    customer_id: requestParam.customer_id,
                    seeker_id: elem.seeker_id,
                    reservation_id: elem.reservation_id
                })
                let seeker_rate = await query.selectWithAndOne(dbConstants.dbSchema.seeker_ratings, {
                    customer_id: requestParam.customer_id,
                    seeker_id: elem.seeker_id,
                    reservation_id: elem.reservation_id
                }, {
                    _id: 0,
                    rating_id: 1
                });
                console.log(seeker_rate)
                if (seeker_rate) {
                    elem.is_rated = true;
                }

                let is_chat = 'read';
                if (elem.is_chat) {
                    _.each(elem.is_chat.chat, (rec) => {
                        if (rec.opponent_user_id == requestParam.customer_id) {
                            if (rec.status.includes(requestParam.customer_id) == false) {
                                is_chat = 'unread'
                            }
                        }
                    })
                }
                elem.is_chat = is_chat

                delete elem.from_date
                delete elem.to_date
                delete elem.parked_date
            })
            resolve(reservation)
            return;
        } catch (error) {
            console.log(error)
            reject(error)
            return
        }
    })
};

/*
    Name : bankAccountTypePaymentList
    Purpose : bankAccountTypePaymentList
    Original Author : Gaurav Patel
    Created At : 06th APr 2021
*/
const bankAccountTypePaymentList = async (requestParam) => {
    return new Promise(async (resolve, reject) => {
        try {

            let banks = await query.selectWithAnd(dbConstants.dbSchema.banks, {
                status: 'active'
            }, {
                _id: 0,
                bank_id: 1,
                title: 1
            }, {
                created_at: 1
            });
            _.each(banks, (element, index, list) => {
                element['title'] = element.title[requestParam.code] || '';
            });

            let bank_account_types = await query.selectWithAnd(dbConstants.dbSchema.bank_account_types, {
                status: 'active'
            }, {
                _id: 0,
                bank_account_type_id: 1,
                title: 1
            }, {
                created_at: 1
            });
            _.each(bank_account_types, (element, index, list) => {
                element['title'] = element.title[requestParam.code] || '';
            });

            let payment_frequencies = await query.selectWithAnd(dbConstants.dbSchema.payment_frequencies, {
                status: 'active'
            }, {
                _id: 0,
                payment_frequency_id: 1,
                title: 1
            }, {
                created_at: 1
            });
            _.each(payment_frequencies, (element, index, list) => {
                element['title'] = element.title[requestParam.code] || '';
            });
            resolve({
                banks: banks,
                bank_account_types: bank_account_types,
                payment_frequencies: payment_frequencies
            })
            return;
        } catch (error) {
            console.log(error)
            reject(error)
            return
        }
    })
};

/*
    Name : bankDetails
    Purpose : bankDetails
    Original Author : Gaurav Patel
    Created At : 06th APr 2021
*/
// const bankDetails = async(requestParam) => {
//     return new Promise(async(resolve, reject) => {
//         try {
//             let response = await query.selectWithAndOne(dbConstants.dbSchema.customers, {
//                 customer_id: requestParam.customer_id
//             }, {
//                 _id: 0,
//                 customer_id: 1,
//                 name: 1,
//                 email: 1,
//                 stripe_profile_id: 1,
//                 bank_details_added: 1,
//                 bank_details: 1
//             });
//             if (!response) {
//                 reject(errors.userNotFound(true, requestParam.code));
//                 return;
//             }

//             let settings = await query.selectWithAndOne(dbConstants.dbSchema.settings, {}, {
//                 _id: 0,
//                 is_payment_live: 1
//             });
//             let stripe = require("stripe")(settings.is_payment_live ? config.stripeInfo.live_key : config.stripeInfo.test_key);

//             // const bankAccount = await stripe.accounts.createExternalAccount(
//             //     'acct_17zKooC79C7G4cmN',
//             //     {
//             //         external_account: 'btok_1J5SyKAEKwRlFi6O8XmR3mEv',
//             //     }
//             // );
//             // console.log(bankAccount)

//             // let payout = await (new Promise(async(resolve1, reject1) => {
//             //     stripe.payouts.create({
//             //         amount: 1100,
//             //         currency: 'usd',
//             //         source_type: 'bank_account',
//             //         destination: 'ba_1JD5EPAEKwRlFi6OIWW1bHDR',
//             //     }, async (error, body) => {
//             //         console.log(error)
//             //         resolve1(body)
//             //     });
//             // }));
//             // console.log(payout);
//             // return false;

//             let stripe_profile_id = response.stripe_profile_id;
//             if (!stripe_profile_id) {
//                 let stripeProfile = await stripe.customers.create({
//                     email: response.email,
//                     name: response.name,
//                     description: `${config.appName} - ${response.name} - ${response.email}`
//                 });
//                 stripe_profile_id = stripeProfile.id;
//                 await query.updateSingle(dbConstants.dbSchema.customers, {
//                     stripe_profile_id
//                 }, {
//                     customer_id: requestParam.customer_id
//                 });
//             }

//             if (response.bank_details_added) {
//                 let deleted = await (new Promise(async(resolve1, reject1) => {
//                     stripe.customers.deleteSource(stripe_profile_id, response.bank_details.stripe_bank_id, async(error, body) => {
//                         resolve1(body)
//                     });
//                 }));
//                 console.log('bank deleted ', deleted)
//             }
//             stripe.tokens.create({
//                 bank_account: {
//                     country: 'US',
//                     currency: 'usd',
//                     account_holder_name: requestParam.bank_account_holder_name,
//                     account_holder_type: 'individual',
//                     routing_number: requestParam.bank_routing_number,
//                     account_number: requestParam.bank_account_number,
//                 },
//             }, (error, stripeToken) => {
//                 if (error) {
//                     reject(errors.stripeError(true, error.message, requestParam.code));
//                     return;
//                 }
//                 stripe.customers.createSource(stripe_profile_id, {
//                     source: stripeToken.id
//                 }, async(error, bankAccount) => {
//                     if (error) {
//                         reject(errors.stripeError(true, error.message, requestParam.code));
//                         return;
//                     }
//                     console.log(bankAccount)
//                     let bank_details = {
//                         bank_account_holder_name: requestParam.bank_account_holder_name,
//                         bank_id: requestParam.bank_id,
//                         bank_account_number: requestParam.bank_account_number,
//                         bank_routing_number: requestParam.bank_routing_number,
//                         bank_account_type_id: requestParam.bank_account_type_id,
//                         payment_frequency_id: requestParam.payment_frequency_id,
//                         stripe_bank_id: bankAccount.id,
//                     }
//                     await query.updateSingle(dbConstants.dbSchema.customers, {
//                         bank_details: bank_details,
//                         bank_details_added: true
//                     }, {
//                         customer_id: requestParam.customer_id
//                     });
//                     resolve({})
//                     return;
//                 });
//             });
//         } catch (error) {
//             console.log(error)
//             reject(error)
//             return
//         }
//     })
// };
const bankDetails = async (requestParam) => {
    return new Promise(async (resolve, reject) => {
        try {
            console.log(requestParam)
            let response = await query.selectWithAndOne(dbConstants.dbSchema.customers, {
                customer_id: requestParam.customer_id
            }, {
                _id: 0,
                customer_id: 1,
                name: 1,
                email: 1,
                stripe_profile_id: 1,
                bank_details_added: 1,
                bank_details: 1,
                stripe_connected_account_id: 1
            });
            if (!response) {
                reject(errors.userNotFound(true, requestParam.code));
                return;
            }
            let settings = await query.selectWithAndOne(dbConstants.dbSchema.settings, {}, { _id: 0, is_payment_live: 1 });
            let stripe = require("stripe")(settings.is_payment_live ? config.stripeInfo.live_key : config.stripeInfo.test_key);
            const plaidClient = new PlaidApi(settings.is_payment_live ? configurationLive : configurationTest);
            // if (requestParam.is_add_bank == true || requestParam.is_add_bank == 'true') {
            if (response.bank_details_added == false && response.stripe_connected_account_id && response.stripe_connected_account_id != '') {
                console.log("========IF========")
                let res = await plaidClient.itemPublicTokenExchange({ public_token: requestParam.plaid_public_token })
                res = await plaidClient.processorStripeBankAccountTokenCreate({ access_token: res.data.access_token, account_id: requestParam.plaid_account_id });
                console.log("\n\nPalid Response : ", res.data)
                const BankTokenID = res.data.stripe_bank_account_token;
                const bankAccount = await stripe.accounts.createExternalAccount(
                    response.stripe_connected_account_id,
                    {
                        external_account: BankTokenID,
                    }
                );

                // To set manual payouts of connected account
                await stripe.accounts.update(
                    response.stripe_connected_account_id,
                    { settings: { payouts: { schedule: { interval: 'manual' } } } }
                );

                //console added by Mukesh Singh
                console.log("\n\n Stripe createExternalAccount Response --> ", bankAccount)
                console.log("--------------------------------------------------------------------------")
                let bank_details = {
                    bank_account_holder_name: requestParam.bank_account_holder_name,
                    // bank_id: requestParam.bank_id,
                    // bank_account_number: requestParam.bank_account_number,
                    // bank_routing_number: requestParam.bank_routing_number,
                    // bank_account_type_id: requestParam.bank_account_type_id,
                    payment_frequency_id: requestParam.payment_frequency_id,
                    bank_accound_id: requestParam.plaid_account_id || '',
                    bank_name: requestParam.bank_name || '',
                    subtype: requestParam.subtype || '',
                    mask: requestParam.mask || '',
                    stripe_bank_id: res.data.stripe_bank_account_token,
                }
                await query.updateSingle(dbConstants.dbSchema.customers, {
                    bank_details: bank_details,
                    bank_details_added: true,
                }, {
                    customer_id: requestParam.customer_id
                });
                resolve({})
                return;
            }
            else {
                console.log("=========Else=========")
                if (response.bank_details_added && (response.stripe_connected_account_id && response.stripe_connected_account_id != '')) {
                    const deleted = await stripe.accounts.del(response.stripe_connected_account_id);
                }
                let accountObj = {
                    type: 'custom',
                    //email: response.email,
                    capabilities: {
                        card_payments: { requested: true },
                        transfers: { requested: true },
                    },
                };
                stripe.accounts.create(accountObj, async (error, account) => {
                    if (error) {
                        reject(errors.stripeError(true, error.message, requestParam.code));
                        return;
                    }

                    //console added by Mukesh Singh
                    console.log("\n\n Stripe Account Details Response --> ", account)
                    console.log("--------------------------------------------------------------------------")
                    const accountLinks = await stripe.accountLinks.create({
                        account: account.id,
                        refresh_url: 'https://backoffice.parksie.com/#/login',
                        return_url: 'https://backoffice.parksie.com/#/login',
                        type: 'account_onboarding',
                    });
                    console.log(accountLinks)
                    let accountLinkURL = "";
                    if (accountLinks && accountLinks.url) {
                        accountLinkURL = accountLinks.url;
                    }
                    await query.updateSingle(dbConstants.dbSchema.customers, {
                        stripe_connected_account_id: account.id,
                        stripe_connected_account_link_url: accountLinkURL,
                        bank_details_added: false

                    }, {
                        customer_id: requestParam.customer_id
                    });
                    resolve({ url: accountLinkURL })
                    return;
                });
            }

            // let res = await plaidClient.itemPublicTokenExchange({public_token:requestParam.plaid_public_token})
            // res = await plaidClient.processorStripeBankAccountTokenCreate({access_token:res.data.access_token, account_id:requestParam.plaid_account_id});
            // let bank_details = {
            //     bank_account_holder_name: requestParam.bank_account_holder_name,
            //     // bank_id: requestParam.bank_id,
            //     // bank_account_number: requestParam.bank_account_number,
            //     // bank_routing_number: requestParam.bank_routing_number,
            //     // bank_account_type_id: requestParam.bank_account_type_id,
            //     payment_frequency_id: requestParam.payment_frequency_id,
            //     bank_accound_id: requestParam.plaid_account_id || '',
            //     bank_name: requestParam.bank_name || '',
            //     subtype: requestParam.subtype || '',
            //     mask: requestParam.mask || '',
            //     stripe_bank_id: res.data.stripe_bank_account_token,
            // }

            // let settings = await query.selectWithAndOne(dbConstants.dbSchema.settings, {}, {_id: 0,is_payment_live: 1});
            // let stripe = require("stripe")(settings.is_payment_live ? config.stripeInfo.live_key : config.stripeInfo.test_key);
            // let accountObj = {
            //     type: 'express',
            //     //email: response.email,
            //     capabilities: {
            //         card_payments: {requested: true},
            //         transfers: {requested: true},
            //     },
            // };
            // stripe.accounts.create(accountObj, async (error, account) => {
            //     if (error) {
            //         reject(errors.stripeError(true, error.message, requestParam.code));
            //         return;
            //     }
            //     const accountLinks = await stripe.accountLinks.create({
            //         account: account.id,
            //         refresh_url: 'http://backoffice.parksie.com/#/login',
            //         return_url: 'http://backoffice.parksie.com/#/login',
            //         type: 'account_onboarding',
            //     });
            //     let accountLinkURL = "";
            //     if (accountLinks && accountLinks.url) {
            //         accountLinkURL = accountLinks.url;
            //     }
            //     await query.updateSingle(dbConstants.dbSchema.customers, {
            //         bank_details: bank_details,
            //         bank_details_added: true,
            //         stripe_connected_account_id: account.id,
            //         stripe_connected_account_link_url: accountLinkURL
            //     }, {
            //         customer_id: requestParam.customer_id
            //     });
            //     resolve({url:accountLinkURL})
            //     return;
            // });
        } catch (error) {
            console.log(error)
            reject(error)
            return
        }
    })
};

const bankDetailsNew = async (requestParam) => {
    return new Promise(async (resolve, reject) => {
        try {
            console.log(requestParam)
            let response = await query.selectWithAndOne(dbConstants.dbSchema.customers, {
                customer_id: requestParam.customer_id
            }, {
                _id: 0,
                customer_id: 1,
                name: 1,
                email: 1,
                stripe_profile_id: 1,
                bank_details_added: 1,
                bank_details: 1,
                stripe_connected_account_id: 1
            });
            if (!response) {
                reject(errors.userNotFound(true, requestParam.code));
                return;
            }
            // let bank_details = {
            //     bank_account_holder_name: requestParam.bank_account_holder_name,

            //     bank_accound_id: requestParam.plaid_account_id || '',
            //     bank_name: requestParam.bank_name || '',
            //     subtype: requestParam.subtype || '',
            //     mask: requestParam.mask || '',
            //     stripe_bank_id: stripe_profile_id,
            // }
            let bank_details = { ...response.bank_details, payment_frequency_id: requestParam.payment_frequency_id || 'PYF161769471815' }
            await query.updateSingle(dbConstants.dbSchema.customers, {
                bank_details: bank_details,
                bank_details_added: true
            }, {
                customer_id: requestParam.customer_id
            });
            resolve({})
            return;
        } catch (error) {
            console.log(error)
            reject(error)
            return
        }
    })
};

const bankDetailsNew2 = async (requestParam) => {
    return new Promise(async (resolve, reject) => {
        try {
            console.log("bankDetailsNew2")
            console.log(requestParam)
            let response = await query.selectWithAndOne(dbConstants.dbSchema.customers, {
                customer_id: requestParam.customer_id
            }, {
                _id: 0,
                customer_id: 1,
                name: 1,
                email: 1,
                stripe_profile_id: 1,
                bank_details_added: 1,
                bank_details: 1,
                stripe_connected_account_id: 1
            });
            if (!response) {
                reject(errors.userNotFound(true, requestParam.code));
                return;
            }
            let settings = await query.selectWithAndOne(dbConstants.dbSchema.settings, {}, { _id: 0, is_payment_live: 1 });
            let stripe = require("stripe")(settings.is_payment_live ? config.stripeInfo.live_key : config.stripeInfo.test_key);
            // const plaidClient = new PlaidApi(settings.is_payment_live ? configurationLive:configurationTest);
            if (response.bank_details_added && (response.stripe_connected_account_id && response.stripe_connected_account_id != '')) {
                const deleted = await stripe.accounts.del(response.stripe_connected_account_id);
            }
            let accountObj = {
                type: 'custom',
                // type: 'custom',
                //email: response.email,
                capabilities: {
                    card_payments: { requested: true },
                    transfers: { requested: true },
                },
            };
            stripe.accounts.create(accountObj, async (error, account) => {
                if (error) {
                    reject(errors.stripeError(true, error.message, requestParam.code));
                    return;
                }
                console.log("============stripe.accounts.create=========")
                console.log(account)
                const accountLinks = await stripe.accountLinks.create({
                    account: account.id,
                    refresh_url: 'https://backoffice.parksie.com/#/login',
                    return_url: 'https://backoffice.parksie.com/#/login',
                    type: 'account_onboarding',
                });
                console.log("=====stripe.accountLinks.create==========")
                console.log(accountLinks)
                let accountLinkURL = "";
                if (accountLinks && accountLinks.url) {
                    accountLinkURL = accountLinks.url;
                }

                // let res = await plaidClient.itemPublicTokenExchange({public_token:requestParam.plaid_public_token})
                // console.log("===itemPublicTokenExchange===")
                // console.log(res.data)
                // res = await plaidClient.processorStripeBankAccountTokenCreate({access_token:res.data.access_token, account_id:requestParam.plaid_account_id});
                // console.log("===processorStripeBankAccountTokenCreate===")
                // console.log(res.data)
                // const BankTokenID = res.data.stripe_bank_account_token;
                // const bankAccount = await stripe.accounts.createExternalAccount(
                //     account.id,
                //     {
                //         external_account: BankTokenID,
                //     }
                // );

                let bank_details = {
                    bank_account_holder_name: requestParam.bank_account_holder_name,
                    // bank_id: requestParam.bank_id,
                    // bank_account_number: requestParam.bank_account_number,
                    // bank_routing_number: requestParam.bank_routing_number,
                    // bank_account_type_id: requestParam.bank_account_type_id,
                    // payment_frequency_id: requestParam.payment_frequency_id,
                    bank_accound_id: requestParam.plaid_account_id || '',
                    bank_name: requestParam.bank_name || '',
                    subtype: requestParam.subtype || '',
                    mask: requestParam.mask || '',
                    // stripe_bank_id: res.data.stripe_bank_account_token,
                }

                await query.updateSingle(dbConstants.dbSchema.customers, {
                    stripe_connected_account_id: account.id,
                    bank_details: bank_details,
                    bank_details_added: true,
                    stripe_connected_account_link_url: accountLinkURL
                }, {
                    customer_id: requestParam.customer_id
                });
                resolve({ url: accountLinkURL })
                return;
            });
        } catch (error) {
            console.log(error)
            reject(error)
            return
        }
    })
};



const stripAccountLink = async (requestParam) => {
    return new Promise(async (resolve, reject) => {
        try {
            console.log(requestParam)
            let customer = await query.selectWithAndOne(dbConstants.dbSchema.customers, { customer_id: requestParam.customer_id }, { customer_id: 1, email: 1 });
            if (!customer) {
                reject(errors.userNotFound(true, requestParam.code));
                return;
            }
            console.log(customer)

            let accountObj = {
                type: 'custom',
                // email: customer.email,
                capabilities: {
                    card_payments: { requested: true },
                    transfers: { requested: true },
                },
            };

            let settings = await query.selectWithAndOne(dbConstants.dbSchema.settings, {}, { _id: 0, is_payment_live: 1 });
            // let stripe = require("stripe")(settings.is_payment_live ? config.stripeInfo.live_key : config.stripeInfo.test_key);
            let stripeKey = (settings.is_payment_live ? config.stripeInfo.live_key : config.stripeInfo.test_key);
            const Stripe = require('stripe');
            const stripe = Stripe(stripeKey);
            let account;
            try {
                account = await stripe.accounts.create(accountObj);
            }
            catch (err) {
                reject(errors.customError(err.raw.message, 409, "Strip error"), null);
                return;
            }

            if (account) {
                const accountLink = await stripe.accountLinks.create({
                    account: account.id,
                    refresh_url: 'https://backoffice.parksie.com/#/login',
                    return_url: 'https://backoffice.parksie.com/#/login',
                    type: 'account_onboarding',
                });

                let updateAdvisorObj = {
                    stripe_connected_account_id: account.id
                }

                let accountLinkURL = "";
                if (accountLink && accountLink.url) {
                    accountLinkURL = accountLink.url;
                    updateAdvisorObj['stripe_connected_account_link_url'] = accountLink.url;
                }

                await query.updateSingle(dbConstants.dbSchema.customers, updateAdvisorObj, { customer_id: requestParam.customer_id })
                resolve({
                    url: accountLinkURL
                });
                return;
            }

        } catch (error) {
            console.log(error)
            reject(error)
            return
        }
    })
};


/*
    Name : Get property list status
    Purpose : Get property list status
    Original Author : Gaurav Patel
    Created At : 13th APr 2021
*/
const propertyStatus = async (requestParam) => {
    return new Promise(async (resolve, reject) => {
        try {
            let response = await query.selectWithAndOne(dbConstants.dbSchema.customers, {
                customer_id: requestParam.customer_id
            }, {
                _id: 0,
                customer_id: 1,
                bank_details: 1
            });
            if (!response) {
                reject(errors.userNotFound(true, requestParam.code));
                return;
            }
            let property = await query.selectWithAndOne(dbConstants.dbSchema.properties, {
                property_id: requestParam.property_id
            }, {
                _id: 0,
                completed_spots: 0,
                created_at: 0,
                updated_at: 0,
                __v: 0
            });
            if (!property) {
                reject(errors.propertyNotFound(true, requestParam.code));
                return;
            }
            let arr = []
            property = JSON.parse(JSON.stringify(property))
            _.each(property.parking_spot_photos, (rec) => {
                if (rec != '') {
                    arr.push(config.aws.prefix + config.aws.s3.propertyBucket + '/' + rec)
                }
            });
            property.parking_spot_photos = arr
            property.bank_details = response.bank_details
            resolve(property)
            return;
        } catch (error) {
            console.log(error)
            reject(error)
            return
        }
    })
};

/*
    Name : Get earning
    Purpose : Get earning
    Original Author : Gaurav Patel
    Created At : 13th APr 2021
*/
const earningGraph = async (requestParam) => {
    return new Promise(async (resolve, reject) => {
        try {
            let settings = await query.selectWithAndOne(dbConstants.dbSchema.settings, {}, {
                _id: 0,
                default_currency: 1
            });
            let response = await query.selectWithAndOne(dbConstants.dbSchema.customers, {
                customer_id: requestParam.customer_id
            }, {
                _id: 0,
                customer_id: 1
            });
            if (!response) {
                reject(errors.userNotFound(true, requestParam.code));
                return;
            }
            let promise = [];
            let matchColumn = {
                owner_id: requestParam.customer_id,
                status: 'completed'
            }
            // if (requestParam.type == 'daily') {
            //     var d = new Date();
            //     var dayName = days[d.getDay()]
            //     let date = timeZone(new Date(d)).tz(requestParam.time_zone).format('YYYY-MM-DD')
            //     matchColumn.completed_date = {
            //         $lte: new Date(date + 'T23:59:59.000Z'),
            //         $gte: new Date(date + 'T00:00:00.000Z')
            //     }
            //     let data = await query.selectWithAnd(dbConstants.dbSchema.reservations, matchColumn, {
            //         _id: 0,
            //         owner_earn: 1,
            //     });
            //     let price = parseFloat(LD.sumBy(data, 'owner_earn')).toFixed(2)
            //     price = parseFloat(price)
            //     let obj = {
            //         label: dayName,
            //         value: price,
            //         date: date,
            //         is_current:true
            //     }
            //     let arr = [obj];
            //     _.each([1,2,3,4,5,6], (elem) => {
            //         var d1 = moment(new Date()).add(elem, 'days');
            //         d1 = new Date(d1)
            //         var dayName = days[d1.getDay()]
            //         arr.push({
            //             label: dayName,
            //             value: 0,
            //             date: timeZone(new Date(d1)).tz(requestParam.time_zone).format('YYYY-MM-DD'),
            //             is_current:false
            //         })
            //     })
            //     resolve({
            //         //graph: [obj],
            //         graph: arr,
            //         payout: {
            //             date: timeZone(new Date(d)).tz(requestParam.time_zone).format('ll'),
            //             amount: price
            //         },
            //         reservation: {
            //             total_number: data.length,
            //             amount: price
            //         },
            //         default_currency: settings.default_currency
            //     })
            //     return;
            // }
            if (requestParam.type == 'daily') {
                let start_date = moment().startOf('week').format('YYYY-MM-DD')
                let end_date = moment().endOf('week').format('YYYY-MM-DD')

                let weekName = []
                function getDates(startDate, endDate) {
                    const dates = []
                    let currentDate = startDate
                    const addDays = function (days) {
                        const date = new Date(this.valueOf())
                        date.setDate(date.getDate() + days)
                        return date
                    }
                    while (currentDate <= endDate) {
                        dates.push(currentDate)
                        weekName.push(days[currentDate.getDay()])
                        currentDate = addDays.call(currentDate, 1)
                    }
                    return dates
                }
                var dateArr = getDates(new Date(start_date), new Date(end_date));
                for (let x in weekName) {
                    promise.push(await dailyEarning(weekName[x], dateArr[x], matchColumn))
                }
                Promise.all(promise)
                    .then(async result => {
                        let total_number = LD.sumBy(result, 'reservation')
                        let amount = LD.sumBy(result, 'value')
                        let payout = await getPayoutAndReservations({ customer_id: requestParam.customer_id, time_zone: requestParam.time_zone })
                        resolve({
                            graph: result,
                            // payout: {date:timeZone(new Date(start_date)).tz(requestParam.time_zone).format('ll')+'-'+timeZone(new Date(end_date)).tz(requestParam.time_zone).format('ll'), amount:amount},
                            // reservation:{total_number:total_number, amount:amount},
                            payout: { date: payout.payout_date, amount: payout.payout_amount },
                            reservation: { total_number: payout.total_res, amount: payout.reservation_amount },
                            default_currency: settings.default_currency
                        });
                        return
                    });
            }
            // else if(requestParam.type == 'weekly'){
            //     let start_date = moment().startOf('week').format('YYYY-MM-DD')
            //     let end_date = moment().endOf('week').format('YYYY-MM-DD')

            //     let weekName = []
            //     var getDaysArray = function(start, end) {
            //         for(var arr=[],dt=timeZone(new Date(start)).tz(requestParam.time_zone); dt<=end; dt.setDate(dt.getDate()+1)){
            //             arr.push(timeZone(new Date(dt)).tz(requestParam.time_zone).format('YYYY-MM-DD'));
            //             weekName.push(days[dt.getDay()])
            //         }
            //         return arr;
            //     };
            //     var dateArr = getDaysArray(new Date(start_date),new Date(end_date));
            //     for(let x in weekName){
            //         promise.push(await weeklyEarning(weekName[x], dateArr[x], matchColumn))
            //     }
            //     Promise.all(promise)
            //     .then(async result => {
            //         let total_number = LD.sumBy(result, 'reservation')
            //         let amount = LD.sumBy(result, 'value')
            //         resolve({
            //             graph: result,
            //             payout: {date:timeZone(new Date(start_date)).tz(requestParam.time_zone).format('ll')+'-'+timeZone(new Date(end_date)).tz(requestParam.time_zone).format('ll'), amount:amount},
            //             reservation:{total_number:total_number, amount:amount},
            //             default_currency: settings.default_currency
            //         });
            //         return
            //     });
            // }
            else if (requestParam.type == 'weekly') {
                let start_date = moment().startOf('month').format('YYYY-MM-DD')
                let end_date = moment().endOf('month').format('YYYY-MM-DD')
                var d = new Date();
                let weeks = await weekCountByMonthYear(d.getMonth(), d.getFullYear());
                for (let x in weeks) {
                    let startDate = new Date(d.getFullYear(), d.getMonth(), weeks[x].start);
                    let endDate = new Date(d.getFullYear(), d.getMonth(), weeks[x].end);
                    promise.push(await weeklyEarning(weeks[x].start + ' - ' + weeks[x].end + ' ' + d.toLocaleString('default', {
                        month: 'short'
                    }), startDate, endDate, matchColumn, requestParam))
                }
                Promise.all(promise)
                    .then(async result => {
                        let total_number = LD.sumBy(result, 'reservation')
                        let amount = LD.sumBy(result, 'value')
                        let payout = await getPayoutAndReservations({ customer_id: requestParam.customer_id, time_zone: requestParam.time_zone })
                        resolve({
                            graph: result,
                            // payout: {
                            //     date: timeZone(new Date(start_date)).tz(requestParam.time_zone).format('ll') + '-' + timeZone(new Date(end_date)).tz(requestParam.time_zone).format('ll'),
                            //     amount: amount
                            // },
                            // reservation: {
                            //     total_number: total_number,
                            //     amount: amount
                            // },
                            payout: { date: payout.payout_date, amount: payout.payout_amount },
                            reservation: { total_number: payout.total_res, amount: payout.reservation_amount },
                            default_currency: settings.default_currency
                        });
                        return
                    });
            } else if (requestParam.type == 'monthly') {
                let start_date = moment().startOf('month').format('YYYY-MM-DD')
                let end_date = moment().endOf('month').format('YYYY-MM-DD')
                for (let x in monthName) {
                    promise.push(await monthlyEarning(monthName[x], x, matchColumn))
                }
                Promise.all(promise)
                    .then(async result => {
                        let total_number = LD.sumBy(result, 'reservation')
                        let amount = LD.sumBy(result, 'value')
                        let payout = await getPayoutAndReservations({ customer_id: requestParam.customer_id, time_zone: requestParam.time_zone })
                        resolve({
                            graph: result,
                            // payout: {
                            //     date: timeZone(new Date(start_date)).tz(requestParam.time_zone).format('ll') + '-' + timeZone(new Date(end_date)).tz(requestParam.time_zone).format('ll'),
                            //     amount: amount
                            // },
                            // reservation: {
                            //     total_number: total_number,
                            //     amount: amount
                            // },
                            payout: { date: payout.payout_date, amount: payout.payout_amount },
                            reservation: { total_number: payout.total_res, amount: payout.reservation_amount },
                            default_currency: settings.default_currency
                        });
                        return
                    });
            } else if (requestParam.type == 'quarterly') {
                let start_date = moment().startOf('year').format('YYYY-MM-DD')
                let end_date = moment().endOf('year').format('YYYY-MM-DD')
                let q = [0, 3, 6, 9];
                for (let x in q) {
                    promise.push(await quarterlyEarning(q[x], matchColumn))
                }
                Promise.all(promise)
                    .then(async result => {
                        let total_number = LD.sumBy(result, 'reservation')
                        let amount = LD.sumBy(result, 'value')
                        let payout = await getPayoutAndReservations({ customer_id: requestParam.customer_id, time_zone: requestParam.time_zone })
                        resolve({
                            graph: result,
                            // payout: {
                            //     date: timeZone(new Date(start_date)).tz(requestParam.time_zone).format('ll') + '-' + timeZone(new Date(end_date)).tz(requestParam.time_zone).format('ll'),
                            //     amount: amount
                            // },
                            // reservation: {
                            //     total_number: total_number,
                            //     amount: amount
                            // },
                            payout: { date: payout.payout_date, amount: payout.payout_amount },
                            reservation: { total_number: payout.total_res, amount: payout.reservation_amount },
                            default_currency: settings.default_currency
                        });
                        return
                    });
            } else if (requestParam.type == 'custom') {
                let finalArr = [];
                var getDaysArray = function (start, end) {
                    for (var arr = [], dt = timeZone(new Date(start)).tz(requestParam.time_zone); dt <= end; dt.setDate(dt.getDate() + 1)) {
                        arr.push(timeZone(new Date(dt)).tz(requestParam.time_zone).format('YYYY-MM-DD'));
                    }
                    return arr;
                };
                var dateArr = getDaysArray(new Date(requestParam.start_date), new Date(requestParam.end_date));
                await forEach(dateArr, async (element) => {
                    var d = new Date(element);
                    let date = timeZone(new Date(d)).tz(requestParam.time_zone).format('YYYY-MM-DD')
                    matchColumn.completed_date = {
                        $lte: new Date(date + 'T23:59:59.000Z'),
                        $gte: new Date(date + 'T00:00:00.000Z')
                    }
                    let data = await query.selectWithAnd(dbConstants.dbSchema.reservations, matchColumn, {
                        _id: 0,
                        owner_earn: 1,
                    });
                    let price = parseFloat(LD.sumBy(data, 'owner_earn')).toFixed(2)
                    price = parseFloat(price)
                    let obj = {
                        label: date,
                        value: price,
                        reservation: data.length
                    }
                    finalArr.push(obj)
                });
                let total_number = LD.sumBy(finalArr, 'reservation')
                let amount = LD.sumBy(finalArr, 'value')
                let payout = await getPayoutAndReservations({ customer_id: requestParam.customer_id, time_zone: requestParam.time_zone })
                resolve({
                    graph: finalArr,
                    // payout: {
                    //     date: timeZone(new Date(requestParam.start_date)).tz(requestParam.time_zone).format('ll') + '-' + timeZone(new Date(requestParam.end_date)).tz(requestParam.time_zone).format('ll'),
                    //     amount: amount
                    // },
                    // reservation: {
                    //     total_number: total_number,
                    //     amount: amount
                    // },
                    payout: { date: payout.payout_date, amount: payout.payout_amount },
                    reservation: { total_number: payout.total_res, amount: payout.reservation_amount },
                    default_currency: settings.default_currency
                })
                return;
            } else {
                resolve([]);
                return
            }
        } catch (error) {
            console.log(error)
            reject(error)
            return
        }
    })
};

const dailyEarning = async (lbl, startDate, matchColumn) => {
    return new Promise(async (resolve, reject) => {
        try {
            let currentDay = moment(new Date()).format('YYYY-MM-DD');

            startDate = moment(new Date(startDate)).format('YYYY-MM-DD')
            matchColumn.completed_date = {
                $lte: new Date(startDate + 'T23:59:59.000Z'),
                $gte: new Date(startDate + 'T00:00:00.000Z')
            }
            let data = await query.selectWithAnd(dbConstants.dbSchema.reservations, matchColumn, {
                _id: 0,
                owner_earn: 1,
            });
            let price = parseFloat(LD.sumBy(data, 'owner_earn')).toFixed(2)
            price = parseFloat(price)
            let is_current = false
            if (startDate == currentDay) {
                is_current = true
            }
            let obj = {
                date: startDate,
                label: lbl,
                value: price,
                reservation: data.length,
                is_current: is_current
            }
            resolve(obj)
            return;
        } catch (error) {
            console.log(error)
            reject(error)
            return
        }
    })
};

const weeklyEarning = async (lbl, startDate, endDate, matchColumn, requestParam) => {
    return new Promise(async (resolve, reject) => {
        try {
            let current_week_start_date = moment().startOf('week').format('YYYY-MM-DD')
            let current_week_end_date = moment().endOf('week').format('YYYY-MM-DD')

            startDate = moment(new Date(startDate)).format('YYYY-MM-DD')
            endDate = moment(new Date(endDate)).format('YYYY-MM-DD')
            matchColumn.completed_date = {
                $lte: new Date(endDate + 'T23:59:59.000Z'),
                $gte: new Date(startDate + 'T00:00:00.000Z')
            }
            let data = await query.selectWithAnd(dbConstants.dbSchema.reservations, matchColumn, {
                _id: 0,
                owner_earn: 1,
            });
            let price = parseFloat(LD.sumBy(data, 'owner_earn')).toFixed(2)
            price = parseFloat(price)
            let is_current = false
            if (startDate >= current_week_start_date && endDate <= current_week_end_date) {
                is_current = true
            }
            let obj = {
                label: lbl,
                value: price,
                reservation: data.length,
                is_current: is_current
            }
            resolve(obj)
            return;
        } catch (error) {
            console.log(error)
            reject(error)
            return
        }
    })
};

const monthlyEarning = async (lbl, index, matchColumn) => {
    return new Promise(async (resolve, reject) => {
        try {
            let currentMonth = new Date();
            currentMonth = currentMonth.toLocaleString('default', { month: 'short' });

            var date = new Date();
            let y = date.getFullYear();
            let m = parseInt(index);
            var firstDay = new Date(y, m, 1);
            firstDay.setHours(0, 0, 0, 0);
            var lastDay = new Date(y, m + 1, 0);
            lastDay.setHours(23, 59, 59, 999);
            matchColumn.completed_date = {
                $lte: lastDay,
                $gte: firstDay
            }
            let data = await query.selectWithAnd(dbConstants.dbSchema.reservations, matchColumn, {
                _id: 0,
                owner_earn: 1,
            });
            let price = parseFloat(LD.sumBy(data, 'owner_earn')).toFixed(2)
            price = parseFloat(price)
            let is_current = false
            if (currentMonth == lbl) {
                is_current = true
            }
            let obj = {
                label: lbl,
                value: price,
                reservation: data.length,
                is_current: is_current
            }
            resolve(obj)
            return;
        } catch (error) {
            console.log(error)
            reject(error)
            return
        }
    })
};

const quarterlyEarning = async (index, matchColumn) => {
    return new Promise(async (resolve, reject) => {
        try {
            let currentMonth = new Date();
            currentMonth = currentMonth.toLocaleString('default', { month: 'short' });

            var date = new Date();
            let y = date.getFullYear();
            let m = parseInt(index);
            var firstDay = new Date(y, m, 1);
            firstDay.setHours(0, 0, 0, 0);
            m = m + 2
            var lastDay = new Date(y, m + 1, 0);
            lastDay.setHours(23, 59, 59, 999);
            matchColumn.completed_date = {
                $lte: lastDay,
                $gte: firstDay
            }
            let data = await query.selectWithAnd(dbConstants.dbSchema.reservations, matchColumn, {
                _id: 0,
                owner_earn: 1,
            });
            let price = parseFloat(LD.sumBy(data, 'owner_earn')).toFixed(2)
            price = parseFloat(price)
            let lbl = 'Jan-Mar';
            if (index == 3) {
                lbl = 'Apr-Jun';
            }
            if (index == 6) {
                lbl = 'Jul-Sep';
            }
            if (index == 9) {
                lbl = 'Oct-Dec';
            }
            let is_current = false
            if (lbl.includes(currentMonth)) {
                is_current = true
            }
            let obj = {
                label: lbl,
                value: price,
                reservation: data.length,
                is_current: is_current
            }
            resolve(obj)
            return;
        } catch (error) {
            console.log(error)
            reject(error)
            return
        }
    })
};

const weekCountByMonthYear = (month, year) => {
    return new Promise(async (resolve, reject) => {
        try {
            var weeks = [],
                firstDate = new Date(year, month, 1),
                lastDate = new Date(year, month + 1, 0),
                numDays = lastDate.getDate();

            var start = 1;
            var end = 7 - firstDate.getDay();
            while (start <= numDays) {
                weeks.push({
                    start: start,
                    end: end
                });
                start = end + 1;
                end = end + 7;
                if (end > numDays)
                    end = numDays;
            }
            resolve(weeks);
            return;
        } catch (error) {
            console.log(error)
            reject(error)
            return
        }
    })
};

const getPayoutAndReservations = (requestParam) => {
    return new Promise(async (resolve, reject) => {
        try {
            let elem = await query.selectWithAndOne(dbConstants.dbSchema.customers, { customer_id: requestParam.customer_id }, { _id: 0, user_type: 1, customer_id: 1, bank_details: 1, stripe_connected_account_id: 1 }, { created_at: -1 });
            let payment_frequency = await query.selectWithAndOne(dbConstants.dbSchema.payment_frequencies, { payment_frequency_id: elem.bank_details.payment_frequency_id }, { _id: 0, payment_frequency_id: 1, title: 1 }, { created_at: -1 });

            let reservations = await query.selectWithAnd(dbConstants.dbSchema.reservations, { status: 'reserved', owner_id: elem.customer_id }, { _id: 0, owner_id: 1, reservation_id: 1, owner_earn: 1 }, { created_at: -1 });
            // let payout_reservations = await query.selectWithAnd(dbConstants.dbSchema.reservations, { status: 'completed', owner_id: elem.customer_id, is_owner_paid: false }, { _id: 0, owner_id: 1, reservation_id: 1, owner_earn: 1 }, { created_at: -1 });
            let nextDate = '';
            if (payment_frequency && payment_frequency.title['EN'] == 'Monthly') {
                var d = new Date();
                d.setMonth(d.getMonth() + 1, 1);
                nextDate = new Date(d)
            }
            if (payment_frequency && payment_frequency.title['EN'] == 'Annually') {
                var d = new Date();
                let yr = d.getFullYear()
                yr = parseFloat(yr)
                nextDate = new Date(yr + 1, 0, 1)
            }
            if (payment_frequency && payment_frequency.title['EN'] == 'Quarterly') {
                var today = new Date();
                var quarter = Math.floor((today.getMonth() + 3) / 3);
                if (quarter == 4) {
                    nextDate = new Date(today.getFullYear() + 1, 0, 1);
                } else {
                    nextDate = new Date(today.getFullYear(), quarter * 3, 1);
                }
            }

            /**
            * Retrieve stripe amount to show on owner dashboard
            */
            let settings = await query.selectWithAndOne(dbConstants.dbSchema.settings, {}, { _id: 0, is_payment_live: 1 });
            let stripe = require("stripe")(settings.is_payment_live ? config.stripeInfo.live_key : config.stripeInfo.test_key);
            const balance = await stripe.balance.retrieve({
                stripeAccount: elem.stripe_connected_account_id,
            });

            let payoutAmount = balance.available[0].amount / 100;
            resolve({
                payout_date: (nextDate !== '') ? timeZone(new Date(nextDate)).tz(requestParam.time_zone).format('ll') : 'Please select payment frequency',
                // payout_amount: LD.sumBy(payout_reservations, 'owner_earn'),
                payout_amount: payoutAmount,
                total_res: reservations.length,
                reservation_amount: LD.sumBy(reservations, 'owner_earn'),
            });
            return;
        } catch (error) {
            console.log(error)
            reject(error)
            return
        }
    })
};

/*
    Name : delete property
    Purpose : delete property
    Original Author : Gaurav Patel
    Created At : 19th APr 2021
*/
const deleteProperty = async (requestParam) => {
    return new Promise(async (resolve, reject) => {
        try {
            let response = await query.selectWithAndOne(dbConstants.dbSchema.customers, {
                customer_id: requestParam.customer_id
            }, {
                _id: 0,
                customer_id: 1
            });
            if (!response) {
                reject(errors.userNotFound(true, requestParam.code));
                return;
            }
            let property = await query.selectWithAndOne(dbConstants.dbSchema.properties, {
                property_id: requestParam.property_id
            }, {
                _id: 0,
                property_id: 1
            });
            if (!property) {
                reject(errors.propertyNotFound(true, requestParam.code));
                return;
            }
            await query.removeMultiple(dbConstants.dbSchema.properties, {
                property_id: requestParam.property_id
            })
            resolve({})
            return;
        } catch (error) {
            console.log(error)
            reject(error)
            return
        }
    })
};

/*
    Name : update property
    Purpose : update property
    Original Author : Gaurav Patel
    Created At : 19th APr 2021
*/
const updateDiscoverableProperty = async (requestParam) => {
    return new Promise(async (resolve, reject) => {
        try {
            let response = await query.selectWithAndOne(dbConstants.dbSchema.customers, {
                customer_id: requestParam.customer_id
            }, {
                _id: 0,
                customer_id: 1
            });
            if (!response) {
                reject(errors.userNotFound(true, requestParam.code));
                return;
            }
            let property = await query.selectWithAndOne(dbConstants.dbSchema.properties, {
                property_id: requestParam.property_id
            }, {
                _id: 0,
                property_id: 1
            });
            if (!property) {
                reject(errors.propertyNotFound(true, requestParam.code));
                return;
            }
            await query.updateSingle(dbConstants.dbSchema.properties, requestParam, {
                property_id: requestParam.property_id
            });
            resolve({})
            return;
        } catch (error) {
            console.log(error)
            reject(error)
            return
        }
    })
};

/*
    Name : list property
    Purpose : list property
    Original Author : Gaurav Patel
    Created At : 19th APr 2021
*/
const propertyList = async (requestParam) => {
    return new Promise(async (resolve, reject) => {
        try {
            let settings = await query.selectWithAndOne(dbConstants.dbSchema.settings, {}, {
                _id: 0,
                default_currency: 1
            });
            let response = await query.selectWithAndOne(dbConstants.dbSchema.customers, {
                customer_id: requestParam.customer_id
            }, {
                _id: 0,
                customer_id: 1
            });
            if (!response) {
                reject(errors.userNotFound(true, requestParam.code));
                return;
            }
            let joinArr = [{
                $lookup: {
                    from: 'reservations',
                    localField: 'property_id',
                    foreignField: 'property_id',
                    as: 'resDetails'
                }
            }, {
                $match: {
                    customer_id: requestParam.customer_id,
                }
            }, {
                $sort: {
                    created_at: -1
                }
            }, {
                $project: {
                    _id: 0,
                    completed_spots: 0,
                    created_at: 0,
                    updated_at: 0,
                    __v: 0,
                }
            }];
            let property = await query.joinWithAnd(dbConstants.dbSchema.properties, joinArr);

            property = JSON.parse(JSON.stringify(property))
            _.each(property, (elem) => {
                let arr = [];
                _.each(elem.parking_spot_photos, (rec) => {
                    if (rec != '') {
                        arr.push(config.aws.prefix + config.aws.s3.propertyBucket + '/' + rec)
                    }
                });
                elem.parking_spot_photos = arr

                let earning = 0;
                let payout = 0;
                _.each(elem.resDetails, (item) => {
                    if (item.status == 'completed') {
                        earning += item.owner_earn ? parseFloat(item.owner_earn) : 0
                        if (item.is_owner_paid == false) {
                            payout += item.owner_earn ? parseFloat(item.owner_earn) : 0
                        }
                    }
                });

                elem.earning = parseFloat(parseFloat(earning).toFixed(2));
                elem.payout = parseFloat(parseFloat(payout).toFixed(2));
                elem.default_currency = settings.default_currency;

                delete elem.resDetails
            });
            resolve(property)
            return;
        } catch (error) {
            console.log(error)
            reject(error)
            return
        }
    })
};

/*
    Name : parking spot list property
    Purpose : parking spot list property
    Original Author : Gaurav Patel
    Created At : 19th May 2021
*/
const parkingSpotList = async (requestParam) => {
    return new Promise(async (resolve, reject) => {
        try {
            let response = await query.selectWithAndOne(dbConstants.dbSchema.customers, {
                customer_id: requestParam.customer_id
            }, {
                _id: 0,
                customer_id: 1
            });
            if (!response) {
                reject(errors.userNotFound(true, requestParam.code));
                return;
            }
            let property = await query.selectWithAnd(dbConstants.dbSchema.properties, {
                customer_id: requestParam.customer_id,
                address: requestParam.address
            }, {
                _id: 0,
                completed_spots: 0,
                created_at: 0,
                updated_at: 0,
                __v: 0
            });
            property = JSON.parse(JSON.stringify(property))
            _.each(property, (elem) => {
                let arr = [];
                _.each(elem.parking_spot_photos, (rec) => {
                    if (rec != '') {
                        arr.push(config.aws.prefix + config.aws.s3.propertyBucket + '/' + rec)
                    }
                });
                elem.parking_spot_photos = arr
            });
            resolve(property)
            return;
        } catch (error) {
            console.log(error)
            reject(error)
            return
        }
    })
};

/*
    Name :  list property
    Purpose : list property
    Original Author : Gaurav Patel
    Created At : 14th Jun 2021
*/
const propertyListWithTotalSpot = async (requestParam) => {
    return new Promise(async (resolve, reject) => {
        try {
            let response = await query.selectWithAndOne(dbConstants.dbSchema.customers, {
                customer_id: requestParam.customer_id
            }, {
                _id: 0,
                customer_id: 1
            });
            if (!response) {
                reject(errors.userNotFound(true, requestParam.code));
                return;
            }
            let property = await query.selectWithAnd(dbConstants.dbSchema.properties, {
                customer_id: requestParam.customer_id
            }, {
                _id: 0,
                completed_spots: 0,
                created_at: 0,
                updated_at: 0,
                __v: 0
            });
            let address = _.pluck(property, 'address');
            address = _.uniq(address);
            let arr = [];
            await Promise.all(address.map(async (elem) => {
                let properties = await query.selectWithAnd(dbConstants.dbSchema.properties, {
                    customer_id: requestParam.customer_id,
                    address: elem
                }, {
                    _id: 0,
                    completed_spots: 0,
                    created_at: 0,
                    updated_at: 0,
                    __v: 0
                });
                arr.push({
                    address: elem,
                    property_type: properties[0].property_type,
                    total_spots: properties.length
                })
            }));
            arr = _.uniq(arr, true, function (item) {
                return item.address;
            });
            resolve(arr)
            return;
        } catch (error) {
            console.log(error)
            reject(error)
            return
        }
    })
};

/*
    Name :  delete property
    Purpose : delete property
    Original Author : Gaurav Patel
    Created At : 14th Jun 2021
*/
const deletePropertyWithAddress = async (requestParam) => {
    return new Promise(async (resolve, reject) => {
        try {
            let response = await query.selectWithAndOne(dbConstants.dbSchema.customers, {
                customer_id: requestParam.customer_id
            }, {
                _id: 0,
                customer_id: 1
            });
            if (!response) {
                reject(errors.userNotFound(true, requestParam.code));
                return;
            }
            await query.removeMultiple(dbConstants.dbSchema.properties, {
                customer_id: requestParam.customer_id,
                address: requestParam.address
            })
            resolve({})
            return;
        } catch (error) {
            console.log(error)
            reject(error)
            return
        }
    })
};

/*
    Name : rating list
    Purpose : rating list
    Original Author : Gaurav Patel
    Created At : 3rd May 2021
*/
const ratingList = async (requestParam) => {
    return new Promise(async (resolve, reject) => {
        try {
            let response = await query.selectWithAndOne(dbConstants.dbSchema.customers, {
                customer_id: requestParam.customer_id
            }, {
                _id: 0,
                customer_id: 1
            });
            if (!response) {
                reject(errors.userNotFound(true, requestParam.code));
                return;
            }
            let joinArr = [{
                $lookup: {
                    from: 'customers',
                    localField: 'customer_id',
                    foreignField: 'customer_id',
                    as: 'cusDetails'
                }
            }, {
                $unwind: "$cusDetails"
            }, {
                $match: {
                    owner_id: requestParam.customer_id,
                }
            }, {
                $sort: {
                    created_at: -1
                }
            }, {
                $project: {
                    _id: 0,
                    rating_id: "$rating_id",
                    name: "$cusDetails.name",
                    profile_picture: "$cusDetails.profile_picture",
                    rating: "$rating",
                    rating_status: "$rating_status",
                    type: "owner",
                }
            }];
            let owner_rating = await query.joinWithAnd(dbConstants.dbSchema.owner_ratings, joinArr);
            _.each(owner_rating, (elem) => {
                elem.profile_picture = elem.profile_picture != '' ? config.aws.prefix + config.aws.s3.customerBucket + '/' + elem.profile_picture : ''
            });

            let property = await query.selectWithAnd(dbConstants.dbSchema.properties, {
                customer_id: requestParam.customer_id
            }, {
                _id: 0,
                customer_id: 1,
                property_id: 1
            });
            joinArr = [{
                $lookup: {
                    from: 'customers',
                    localField: 'customer_id',
                    foreignField: 'customer_id',
                    as: 'cusDetails'
                }
            }, {
                $unwind: "$cusDetails"
            }, {
                $match: {
                    property_id: {
                        $in: _.pluck(property, 'property_id')
                    },
                }
            }, {
                $sort: {
                    created_at: -1
                }
            }, {
                $project: {
                    _id: 0,
                    rating_id: "$property_rating_id",
                    name: "$cusDetails.name",
                    profile_picture: "$cusDetails.profile_picture",
                    rating: "$rating",
                    rating_status: "$rating_status",
                    type: "property",
                }
            }];
            let property_rating = await query.joinWithAnd(dbConstants.dbSchema.property_ratings, joinArr);
            _.each(property_rating, (elem) => {
                elem.profile_picture = elem.profile_picture != '' ? config.aws.prefix + config.aws.s3.customerBucket + '/' + elem.profile_picture : ''
            });
            owner_rating.push(property_rating)

            resolve(_.flatten(owner_rating))
            return;
        } catch (error) {
            console.log(error)
            reject(error)
            return
        }
    })
};

/*
    Name : seeker-rating
    Purpose : seeker-rating
    Original Author : Gaurav Patel
    Created At : 03rd May 2021
*/
const seekerRating = async (requestParam) => {
    return new Promise(async (resolve, reject) => {
        try {
            let response = await query.selectWithAndOne(dbConstants.dbSchema.customers, {
                customer_id: requestParam.customer_id
            }, {
                _id: 0,
                customer_id: 1
            });
            if (!response) {
                reject(errors.userNotFound(true, requestParam.code));
                return;
            }
            let seeker = await query.selectWithAndOne(dbConstants.dbSchema.customers, {
                customer_id: requestParam.seeker_id
            }, {
                _id: 0,
                customer_id: 1
            });
            if (!seeker) {
                reject(errors.userNotFound(true, requestParam.code));
                return;
            }
            await query.insertSingle(dbConstants.dbSchema.seeker_ratings, requestParam);
            resolve({})
            return;
        } catch (error) {
            console.log(error)
            reject(error)
            return
        }
    })
};

/*
    Name :  Create Plaid Token
    Purpose : Create Plaid Token
    Original Author : Gaurav Patel
    Created At : 02nd Sep 2021
*/
const plaidToken = async (requestParam) => {
    return new Promise(async (resolve, reject) => {
        try {
            let settings = await query.selectWithAndOne(dbConstants.dbSchema.settings, {}, { _id: 0, is_payment_live: 1 });
            const plaidClient = new PlaidApi(settings.is_payment_live ? configurationLive : configurationTest);
            const response = await plaidClient.linkTokenCreate({
                user: {
                    client_user_id: requestParam.customer_id,
                },
                client_name: requestParam.client_name,
                products: ['auth'],
                country_codes: ['US'],
                language: 'en',
            });
            resolve({ token: response.data.link_token })
            return;
        } catch (error) {
            console.log(error)
            reject(error)
            return
        }
    })
};

/*
    Name :  Cron for overage feed
    Purpose : Cron for overage feed
    Original Author : Gaurav Patel
    Created At : 29nd Sep 2021
*/
const cronForOverageFeed = async (requestParam) => {
    return new Promise(async (resolve, reject) => {
        try {
            let columnAndValue = {};
            columnAndValue = {
                $or: [{
                    status: 'parked'
                }]
            };
            let data = await query.selectWithAnd(dbConstants.dbSchema.reservations, columnAndValue, { _id: 0 });
            await forEach(data, async (elem) => {
                let todayDate = timeZone(new Date()).tz(requestParam.time_zone);
                let to_date = timeZone(new Date(elem.to_date)).tz(requestParam.time_zone);
                const minutes = todayDate.diff(to_date, 'minutes');
                if (minutes == 16) {
                    let owner = await query.selectWithAndOne(dbConstants.dbSchema.customers, { customer_id: elem.owner_id }, { _id: 0, device_token: 1 });
                    let template = await query.selectWithAndOne(dbConstants.dbSchema.push_templates, { code: 'OWNER_TO_START_OVERAGE_FEE' }, { _id: 0, value: 1, caption_value: 1 }, { created_at: 1 });
                    if (template) {
                        let msg = template.value[requestParam.code];
                        notificationHandler.sendNotification({
                            message: msg,
                            type: 'reservation_overage',
                            title: config.appName,
                            user_id: elem.owner_id,
                            device_token: owner ? [owner.device_token] : [],
                            sender_id: elem.owner_id,
                            reservation_id: elem.reservation_id,
                            user_type: 'owner',
                            reservation_status: 'parked',
                        });
                        notificationHandler.setLogs({
                            user_id: [elem.owner_id],
                            user_type: 'owner',
                            message: template.caption_value[requestParam.code],
                            description: msg,
                            type: 'reservation_overage',
                            address: '',
                            sender_id: elem.owner_id,
                            reservation_id: elem.reservation_id,
                            reservation_status: 'parked',
                        });
                    }
                }
            });
            resolve({})
            return;
        } catch (error) {
            console.log(error)
            reject(error)
            return
        }
    })
};

const createStripeAcc = async (requestParam) => {
    return new Promise(async (resolve, reject) => {
        try {
            console.log("createStripeAcc")
            console.log(requestParam)
            let response = await query.selectWithAndOne(dbConstants.dbSchema.customers, {
                customer_id: requestParam.customer_id
            }, {
                _id: 0,
                customer_id: 1,
                name: 1,
                email: 1,
                stripe_profile_id: 1,
                bank_details_added: 1,
                bank_details: 1,
                stripe_connected_account_id: 1
            });
            if (!response) {
                reject(errors.userNotFound(true, requestParam.code));
                return;
            }
            let settings = await query.selectWithAndOne(dbConstants.dbSchema.settings, {}, { _id: 0, is_payment_live: 1 });
            let stripe = require("stripe")(settings.is_payment_live ? config.stripeInfo.live_key : config.stripeInfo.test_key);
            const plaidClient = new PlaidApi(settings.is_payment_live ? configurationLive : configurationTest);
            if (response.bank_details_added && (response.stripe_connected_account_id && response.stripe_connected_account_id != '')) {
                const deleted = await stripe.accounts.del(response.stripe_connected_account_id);
            }
            let accountObj = {
                type: 'custom',
                capabilities: {
                    card_payments: { requested: true },
                    transfers: { requested: true },
                },
                settings: { payouts: { schedule: { interval: 'manual' } } }
            };
            stripe.accounts.create(accountObj, async (error, account) => {
                if (error) {
                    reject(errors.stripeError(true, error.message, requestParam.code));
                    return;
                }
                console.log("============stripe.accounts.create=========")
                console.log(account)
                const accountLinks = await stripe.accountLinks.create({
                    account: account.id,
                    refresh_url: 'https://backoffice.parksie.com/#/login',
                    return_url: 'https://backoffice.parksie.com/#/login',
                    type: 'account_onboarding',
                });
                console.log("=====stripe.accountLinks.create==========")
                console.log(accountLinks)
                let accountLinkURL = "";
                if (accountLinks && accountLinks.url) {
                    accountLinkURL = accountLinks.url;
                }
                await query.updateSingle(dbConstants.dbSchema.customers, {
                    stripe_connected_account_id: account.id,
                    bank_details_added: true,
                    stripe_connected_account_link_url: accountLinkURL
                }, {
                    customer_id: requestParam.customer_id
                });
                resolve({ url: accountLinkURL })
                return;
            });
        } catch (error) {
            console.log(error)
            reject(error)
            return
        }
    })
};

const connectStripePlaidAcc = async (requestParam) => {
    return new Promise(async (resolve, reject) => {
        try {
            console.log("======connectStripePlaidAcc=======")
            console.log(requestParam)
            let response = await query.selectWithAndOne(dbConstants.dbSchema.customers, {
                customer_id: requestParam.customer_id
            }, {
                _id: 0,
                customer_id: 1,
                name: 1,
                stripe_profile_id: 1,
                stripe_connected_account_id: 1
            });
            console.log("=====response======")
            console.log(response)
            if (!response) {
                reject(errors.userNotFound(true, requestParam.code));
                return;
            }
            if (response.stripe_connected_account_id && response.stripe_connected_account_id !== '') {
                let settings = await query.selectWithAndOne(dbConstants.dbSchema.settings, {}, { _id: 0, is_payment_live: 1 });
                let stripe = require("stripe")(settings.is_payment_live ? config.stripeInfo.live_key : config.stripeInfo.test_key);
                const plaidClient = new PlaidApi(settings.is_payment_live ? configurationLive : configurationTest);
                let res = await plaidClient.itemPublicTokenExchange({ public_token: requestParam.plaid_public_token })
                console.log("===itemPublicTokenExchange===")
                console.log(res.data)
                res = await plaidClient.processorStripeBankAccountTokenCreate({ access_token: res.data.access_token, account_id: requestParam.plaid_account_id });
                console.log("===processorStripeBankAccountTokenCreate===")
                console.log(res.data)
                const BankTokenID = res.data.stripe_bank_account_token;
                const bankAccount = await stripe.accounts.createExternalAccount(
                    response.stripe_connected_account_id,
                    {
                        external_account: BankTokenID,
                    }
                );

                // To set manual payouts of connected account
                await stripe.accounts.update(
                    response.stripe_connected_account_id,
                    { settings: { payouts: { schedule: { interval: 'manual' } } } }
                );
                resolve(bankAccount)
                return;
            } else {
                reject(errors.customError('Stripe Account Not Exist', 404, 'Stripe Account Not Exist', true));
                return;
            }
        } catch (error) {
            console.log(error)
            reject(error)
            return
        }
    })
};


module.exports = {
    reservationList,
    bankAccountTypePaymentList,
    bankDetails,
    bankDetailsNew,
    bankDetailsNew2,
    propertyStatus,
    earningGraph,
    deleteProperty,
    updateDiscoverableProperty,
    propertyList,
    parkingSpotList,
    ratingList,
    seekerRating,
    propertyListWithTotalSpot,
    deletePropertyWithAddress,
    plaidToken,
    cronForOverageFeed,
    getPayoutAndReservations,
    stripAccountLink,
    createStripeAcc,
    connectStripePlaidAcc
};
