'use strict';

const config = require('./../../config');
const logger = require('./../../utils/logger');
const errors = require('./../../utils/dz-errors-api');
const dbConstants = require('./../../constants/db-constants');
const query = require('./../../utils/query-creator-api');
let async = require('async');
let _ = require('underscore');
const commonHandler = require('./../../model_handlers/api/common-handler');
const idGenerator = require('./../../utils/id-generator');
const timeZone = require('moment-timezone');

/*
    Name : Upload or Delete Package Image
    Purpose : Upload or Delete Package Image
    Original Author : Gaurav Patel
    Created At : 17th Mar 2021
*/
const uploadImage = async(requestParam, req) => {
    return new Promise(async(resolve, reject) => {
        try {
            if (requestParam.action == 'upload') {
                let photo = await new Promise((solve, reject) => {
                    commonHandler.uploadImage(req.files.photo, config.aws.s3.propertyBucket , (error, path) => {
                        solve(path)
                    });
                });
                resolve(photo);
                return;
            } else {
                let fileObjects = [{
                    Key: 'parksie/properties/' + /[^/]*$/.exec(requestParam.photo)[0]
                }];
                commonHandler.removeMultipleImages(fileObjects, (error, path) => {
                    resolve({});
                    return;
                });
            }
        } catch (error) {
            console.log(error);
            reject(errors.internalServer(true, requestParam.code));
            return;
        }
    })
};

/*
    Name : list
    Purpose : list
    Original Author : Gaurav Patel
    Created At : 12th March 2021
*/
const list = async(requestParam, req) => {
    return new Promise(async(resolve, reject) => {
        try {
            let settings = await query.selectWithAndOne(dbConstants.dbSchema.settings, {}, { _id: 0, default_currency: 1} );
            
            let response = await query.selectWithAndOne(dbConstants.dbSchema.customers, {customer_id: requestParam.customer_id}, { _id: 0, customer_id: 1} );
            if(!response){
                reject(errors.userNotFound(true, requestParam.code));
                return;
            }
            let joinArr = [{
                $lookup: {
                    from: 'reservations',
                    localField: 'property_id',
                    foreignField: 'property_id',
                    as: 'resDetails'
                }
            }, {
                $match: {
                    customer_id: requestParam.customer_id,
                }
            }, {
                $sort: {
                    created_at: -1
                }
            }, {
                $project: {
                    _id: 0,
                    completed_spots: 0,
                    created_at: 0,
                    updated_at: 0,
                    __v: 0,
                }
            }];
            let list = await query.joinWithAnd(dbConstants.dbSchema.properties, joinArr);
            //let list = await query.selectWithAnd(dbConstants.dbSchema.properties, {customer_id: requestParam.customer_id}, { _id: 0, completed_spots:0, created_at: 0, updated_at:0, __v:0}, { created_at: -1} );
            let arr = [];
            _.each(list, (elem) => {
                _.each(elem.parking_spot_photos, (rec) => {
                    if(rec != ''){
                        arr.push(config.aws.prefix + config.aws.s3.propertyBucket + '/' + rec)
                    }
                });
                elem.parking_spot_photos = arr

                let earning = 0;
                _.each(elem.resDetails, (item) => {
                    if(item.status == 'completed'){
                        earning += item.owner_earn ? parseFloat(item.owner_earn) : 0
                    }
                });

                elem.earning = parseFloat(parseFloat(earning).toFixed(2));
                elem.default_currency = settings.default_currency;

                delete elem.resDetails
            });
            resolve(list)
            return;
        } catch (error) {
            reject(error)
            return
        }
    })
};


/*
    Name : add
    Purpose : add
    Original Author : Gaurav Patel
    Created At : 12th March 2021
*/
const add = async(requestParam, req) => {
    return new Promise(async(resolve, reject) => {
        try {
            let response = await query.selectWithAndOne(dbConstants.dbSchema.customers, {customer_id: requestParam.customer_id}, { _id: 0, customer_id: 1} );
            if(!response){
                reject(errors.userNotFound(true, requestParam.code));
                return;
            }
            requestParam.contact_details = JSON.parse(requestParam.contact_details);

            requestParam.location = {
                type: "Point",
                coordinates: [requestParam.longitude, requestParam.latitude],
            };

            await Promise.all(_.range(parseFloat(requestParam.parking_spots)).map(async (element) => {
                let property = await query.insertSingle(dbConstants.dbSchema.properties, requestParam);
            }));
            resolve({})
            return;
        } catch (error) {
            console.log(error)
            reject(error)
            return
        }
    })
};

/*
    Name : update property
    Purpose : update property
    Original Author : Gaurav Patel
    Created At : 12th March 2021
*/
const update = async(requestParam, req) => {
    return new Promise(async(resolve, reject) => {
        try {
            let response = await query.selectWithAndOne(dbConstants.dbSchema.customers, {customer_id: requestParam.customer_id}, { _id: 0, customer_id: 1} );
            if(!response){
                reject(errors.userNotFound(true, requestParam.code));
                return;
            }
            let property = await query.selectWithAndOne(dbConstants.dbSchema.properties, {property_id: requestParam.property_id}, { _id: 0, property_id: 1, completed_spots:1} );
            if(!property){
                reject(errors.propertyNotFound(true, requestParam.code));
                return;
            }
            let completed_spots = property.completed_spots;

            let obj = {}
            if(requestParam.method == 'geo_tag'){
                obj.geo_tag = {
                    latitude: parseFloat(requestParam.latitude),
                    longitude: parseFloat(requestParam.longitude),
                    altitude: requestParam.altitude ? parseFloat(requestParam.altitude) : 0,
                    address: requestParam.address,
                }
                obj.is_added_geo_tag = true;
                completed_spots += 1
            }

            if(requestParam.method == 'parking_size'){
                obj.parking_size = {
                    vehicle_type_id: requestParam.vehicle_type_id,
                    is_apply: (requestParam.is_apply == 'true' ? true : false) || false,
                    width: requestParam.width || 0,
                    height: requestParam.height || 0,
                    length: requestParam.length || 0,
                }
                obj.is_added_parking_size = true;
                completed_spots += 1
            }

            if(requestParam.method == 'amenities'){
                obj.parking_spot_type_id = requestParam.parking_spot_type_id;
                obj.is_apply_parking_spot_type = requestParam.is_apply_parking_spot_type == 'true' ? true : false;

                obj.parking_spot_coverage_id = requestParam.parking_spot_coverage_id;
                obj.is_apply_parking_spot_coverage = requestParam.is_apply_parking_spot_coverage == 'true' ? true : false;

                obj.parking_spot_terrain_id = requestParam.parking_spot_terrain_id;
                obj.is_apply_parking_spot_terrain = requestParam.is_apply_parking_spot_terrain == 'true' ? true : false;

                obj.parking_spot_amenity_id = obj.parking_spot_amenity_id ? requestParam.parking_spot_amenity_id.split(',') : [];
                if(requestParam.is_apply_parking_spot_amenity){
                    obj.is_apply_parking_spot_amenity = requestParam.is_apply_parking_spot_amenity == 'true' ? true : false;
                }

                obj.is_added_amenities = true;
                completed_spots += 1

            }

            if(requestParam.method == 'description_and_photos'){
                obj.parking_spot_description = requestParam.parking_spot_description;
                obj.is_apply_parking_spot_description = requestParam.is_apply_parking_spot_description == 'true' ? true : false;
                obj.parking_spot_photos = requestParam.parking_spot_photos.split(',');
                obj.is_apply_parking_spot_photos = requestParam.is_apply_parking_spot_photos == 'true' ? true : false;
                    
                obj.is_added_description_and_photos = true;
                completed_spots += 1
            }

            if(requestParam.method == 'availability'){
                obj.is_specific_date_range = requestParam.is_specific_date_range == 'true' ? true : false;
                obj.from_date_range = requestParam.from_date_range || timeZone(new Date()).tz(requestParam.time_zone);
                obj.to_date_range = requestParam.to_date_range || timeZone(new Date()).tz(requestParam.time_zone);

                obj.available_days = requestParam.available_days.split(',')

                obj.from_reservation_time = requestParam.from_reservation_time
                obj.to_reservation_time = requestParam.to_reservation_time

                obj.from_entry_time = requestParam.from_entry_time
                obj.to_entry_time = requestParam.to_entry_time

                obj.is_auto_activate = requestParam.is_auto_activate
                obj.distance_from_parking_spot = parseFloat(requestParam.distance_from_parking_spot) || 0

                obj.is_parking_availability = requestParam.is_parking_availability == 'true' ? true : false

                obj.grace_period = requestParam.grace_period
                obj.is_added_grace_period = true;

                obj.is_added_availability = true;
                completed_spots += 1
            }

            if(requestParam.method == 'access_instruction'){
                obj.access_instruction_id = requestParam.access_instruction_id;
                obj.private_instruction = requestParam.private_instruction || '';
                obj.is_apply_access_instruction = requestParam.is_apply_access_instruction == 'true' ? true : false;

                obj.is_added_access_instruction = true;
                completed_spots += 1
            }

            if(requestParam.method == 'pricing'){
                obj.pricing = JSON.parse(requestParam.pricing);
                obj.is_apply_pricing = requestParam.is_apply_pricing == 'true' ? true : false;

                obj.is_added_pricing = true;
                completed_spots += 1
            }
            
            obj.completed_spots = completed_spots
            
            if(completed_spots == 7){
                obj.status = 'completed'
            }

            await query.updateSingle(dbConstants.dbSchema.properties, obj, {property_id: requestParam.property_id});
            resolve({})
            return;
        } catch (error) {
            console.log(error)
            reject(error)
            return
        }
    })
};

module.exports = {
    list,
    add,
    update,
    uploadImage
};