'use strict';

const logger = require('./../../utils/logger');
const jsonResponse = require('./../../utils/json-response');
const errors = require('./../../utils/dz-errors');
const dbConstants = require('./../../constants/db-constants');
const config = require('./../../config');
const query = require('./../../utils/query-creator');
const queryApi = require('./../../utils/query-creator-api');
let asyncLoop = require('async');
let _ = require('underscore');
let moment = require('moment');
const commonHandler = require('./../../model_handlers/backend/common-handler');
const emailHandler = require('./../../model_handlers/api/email-handler');
const passwordHandler = require('./../../utils/password');
const { forEach } = require('p-iteration');
const timeZone = require('moment-timezone');
var days = ['Sun', 'Mon', 'Tue', 'Wed', 'Thu', 'Fri', 'Sat'];
let monthName = ["Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"];
const LD = require('lodash');

const get = async function (req, done) {
    let columnAndValue = {}
    if (req.query.customer_id || req.query.status || req.query.user_type) {
        if (req.query.customer_id) {
            columnAndValue.customer_id = req.query.customer_id;
        }
        if (req.query.status) {
            columnAndValue.status = req.query.status;
        }

        if (req.query.user_type) {
            columnAndValue.user_type = { $in: req.query.user_type.split(',') };
            columnAndValue.status = 'active';
        }
        query.selectWithAndFilter(dbConstants.dbSchema.customers, columnAndValue, {
            _id: 0,
        }, { created_at: -1 }, {}, (error, response) => {
            if (error) {
                logger('Error: can not get ', dbConstants.dbSchema.customers);
                done(errors.internalServer(true), null);
                return;
            }
            if (req.query.status || req.query.user_type) {
                done(null, response)
            }
            else {
                response = response[0]
                response = JSON.parse(JSON.stringify(response))
                query.selectWithAndOne(dbConstants.dbSchema.languages, { language_id: response.language_id }, {
                    title: 1
                }, async (error, language) => {
                    if (response.profile_picture != '') {
                        response.profile_picture = config.aws.prefix + config.aws.s3.customerBucket + '/' + response.profile_picture;
                    }
                    response.language = language ? language.title : ''
                    done(null, response)
                });
            }
        });
    }
    else {
        if (req.query.is_dashboard) {
            let date = timeZone(new Date()).tz(req.headers.time_zone).format('YYYY-MM-DD');
            columnAndValue.created_at = {
                $lt: new Date(date + 'T23:59:59.000Z'),
                $gte: new Date(date + 'T00:00:00.000Z')
            }
        }
        if (req.query.signup_with) {
            if (req.query.signup_with != '') {
                columnAndValue.register_type = req.query.signup_with
            }
        }

        let joinArr = [{
            $lookup: {
                from: 'owner_ratings',
                localField: 'customer_id',
                foreignField: 'owner_id',
                as: 'ratings'
            }
        }, {
            $match: columnAndValue
        }, {
            $sort: {
                created_at: -1
            }
        }, {
            $project: {
                _id: 0,
                customer_id: 1,
                name: 1,
                surname: 1,
                mobile_country_code: 1,
                mobile: 1,
                email: 1,
                created_at: 1,
                last_login: 1,
                register_type: 1,
                device_type: 1,
                status: 1,
                user_type: 1,
                ratings: "$ratings"
            }
        }];
        let response = await queryApi.joinWithAnd(dbConstants.dbSchema.customers, joinArr);
        response = JSON.parse(JSON.stringify(response))
        _.each(response, (elem) => {
            elem.created_at = timeZone(new Date(elem.created_at)).tz(req.headers.time_zone).format('lll');
            if (elem.last_login) {
                elem.last_login = timeZone(new Date(elem.last_login)).tz(req.headers.time_zone).format('lll');
            }
            else {
                elem.last_login = ''
            }
            if (elem.ratings.length > 0) {
                let rat = 0;
                _.each(elem.ratings, (rec) => {
                    rat += rec.rating
                });
                rat = rat / elem.ratings.length
                elem.ratings = parseFloat(rat).toFixed(1);
            }
            else {
                elem.ratings = 0;
            }
            if (!elem.user_type) elem.user_type = ''
        });
        done(null, response)
    }
};

const create = async function (requestParam, req, done) {
    let compareColumnAndValues = {
        mobile: requestParam.mobile,
        mobile_country_code: requestParam.mobile_country_code
    };
    query.selectWithAndOne(dbConstants.dbSchema.customers, compareColumnAndValues, {
        customer_id: 1
    }, async (error, exists) => {
        if (error) {
            done(errors.internalServer(true));
            return;
        }
        if (exists) {
            done(errors.duplicateUser(true), null);
            return;
        }
        if (req.files) {
            if (req.files.profile_picture) {
                requestParam.profile_picture = await new Promise((resolve, reject) => {
                    commonHandler.uploadImage(req.files.profile_picture, config.aws.s3.customerBucket, (error, path) => {
                        resolve(path)
                    });
                });
            }
        }
        requestParam.old_password = requestParam.password
        let encryptPassword = await passwordHandler.encrypt(requestParam.password.toString());
        requestParam.password = encryptPassword;

        let measurement = await queryApi.selectWithAndOne(dbConstants.dbSchema.measurements, {}, { _id: 0, measurement_id: 1, title: 1 });
        requestParam.measurement_id = measurement ? measurement.measurement_id : ''

        query.insertSingle(dbConstants.dbSchema.customers, requestParam, function (error, user) {
            if (error) {
                logger('Error: can not create user');
                done(error, null);
                return;
            }
            // requestParam.template_code = 'CUS_SIGN_UP';
            // requestParam.code = 'EN';
            // emailHandler.sendEmail(requestParam);
            done(null, user);
        });
    });
};

const update = function (requestParam, req, done) {
    let compareColumnAndValues = {
        $and: [{
            $or: [{
                mobile: requestParam.mobile,
                mobile_country_code: requestParam.mobile_country_code,
            }]
        }, {
            customer_id: {
                $ne: requestParam.customer_id
            }
        }]

    };
    query.selectWithAndOne(dbConstants.dbSchema.customers, compareColumnAndValues, {
        customer_id: 1,
    }, async (error, exists) => {
        if (error) {
            done(errors.internalServer(true));
            return;
        }
        if (exists) {
            done(errors.duplicateUser(true), null);
            return;
        }
        if (requestParam.change_logo) {
            query.selectWithAndOne(dbConstants.dbSchema.customers, {
                customer_id: requestParam.customer_id
            }, {
                customer_id: 1,
                profile_picture: 1,
            }, async (error, user) => {
                let fileObjects = [];
                if (requestParam.change_logo) {
                    fileObjects.push({
                        Key: 'parksie/customers/' + /[^/]*$/.exec(user.profile_picture)[0]
                    })
                }
                commonHandler.removeMultipleImages(fileObjects, async (error, path) => {
                    requestParam.profile_picture = await new Promise((resolve, reject) => {
                        commonHandler.uploadImage(req.files.profile_picture, config.aws.s3.customerBucket, (error, path) => {
                            resolve(path)
                        });
                    });
                    query.updateSingle(dbConstants.dbSchema.customers, requestParam, {
                        'customer_id': requestParam.customer_id
                    }, function (error, user) {
                        if (error) {
                            logger('Error: can not update push Notification');
                            done(error, null);
                            return;
                        }
                        done(null, user);
                    });
                });
            });
        }
        else {
            delete requestParam.profile_picture
            query.updateSingle(dbConstants.dbSchema.customers, requestParam, {
                'customer_id': requestParam.customer_id
            }, function (error, user) {
                if (error) {
                    logger('Error: can not update push Notification');
                    done(error, null);
                    return;
                }
                done(null, user);
            });
        }
    });
};

const action = (requestParam, done) => {
    if (requestParam['type'] == "delete") {
        query.selectWithAnd(dbConstants.dbSchema.customers, {
            customer_id: {
                $in: requestParam.ids
            }
        }, {
            _id: 0,
            customer_id: 1,
            profile_picture: 1,
        }, (error, user) => {
            let fileObjects = [];
            for (var i = 0; i < user.length; i++) {
                fileObjects.push({
                    Key: 'parksie/customers/' + /[^/]*$/.exec(user[i].profile_picture)[0]
                })
            }
            commonHandler.removeMultipleImages(fileObjects, (error, path) => {
                query.removeMultiple(dbConstants.dbSchema.customers, {
                    'customer_id': {
                        $in: requestParam['ids']
                    }
                }, function (error, data) {
                    if (error) {
                        logger('Error: can not delete ');
                        done(error, null);
                        return;
                    }
                    var userRef = config.firebase.firebaseDBRef.ref('users');
                    let ids = requestParam.ids
                    var newData = {};
                    for (let i = 0; i < ids.length; i++) {
                        var ref = userRef.child(ids[i]);
                        ref.once("value").then(function(snapshot) {
                            if (snapshot.exists()) {
                                newData[ids[i] + "/status"] = requestParam.type;
                                userRef.update(newData);
                            }
                        });
                    }
                    done(null, data);
                });
            });
        });
    }
    else {
        let columnsToUpdate = {
            status: requestParam['type']
        };
        query.updateMultiple(dbConstants.dbSchema.customers, columnsToUpdate, {
            'customer_id': {
                $in: requestParam['ids']
            }
        }, function (error, data) {
            if (error) {
                logger('Error: can not update ');
                done(error, null);
                return;
            }
            var userRef = config.firebase.firebaseDBRef.ref('users');
            let ids = requestParam.ids
            var newData = {};
            for (let i = 0; i < ids.length; i++) {
                var ref = userRef.child(ids[i]);
                ref.once("value").then(function (snapshot) {
                    if (snapshot.exists()) {
                        newData[ids[i] + "/status"] = requestParam.type;
                        userRef.update(newData);
                    }
                });
            }
            done(null, data);
        });
    }
};

const getBankDetails = async (requestParam, done) => {
    let joinArr = [{
        $lookup: {
            from: 'banks',
            localField: 'bank_details.bank_id',
            foreignField: 'bank_id',
            as: 'bankDetails'
        }
    }, {
        $unwind: "$bankDetails"
    }, {
        $match: requestParam
    }, {
        $sort: { created_at: -1 }
    }, {
        $project: {
            _id: 0,
            customer_id: "$customer_id",
            bank_account_holder_name: "$bank_details.bank_account_holder_name",
            bank_account_number: "$bank_details.bank_account_number",
            bank_routing_number: "$bank_details.bank_routing_number",
            bank_name: "$bankDetails.title.EN",
        }
    }];
    query.joinWithAnd(dbConstants.dbSchema.customers, joinArr, async (error, response) => {
        if (error) {
            logger('Error: can not get record.');
            done(errors.internalServer(true), null);
            return;
        }
        done(null, response)
    });
}

const getPropertyList = async (requestParam, done) => {
    let joinArr = [{
        $lookup: {
            from: 'parking_spot_coverages',
            localField: 'parking_spot_coverage_id',
            foreignField: 'parking_spot_coverage_id',
            as: 'spotCoverageDetails'
        }
    }, {
        $unwind: "$spotCoverageDetails"
    }, {
        $lookup: {
            from: 'parking_spot_types',
            localField: 'parking_spot_type_id',
            foreignField: 'parking_spot_type_id',
            as: 'spotTypesDetails'
        }
    }, {
        $unwind: "$spotTypesDetails"
    }, {
        $match: requestParam
    }, {
        $sort: { created_at: -1 }
    }, {
        $project: {
            _id: 0,
            property_id: 1,
            address: 1,
            parking_spot_photos: 1,
            pricing: 1,
            status: 1,
            parking_spot_type_title: "$spotTypesDetails.title.EN",
            parking_spot_coverage_title: "$spotCoverageDetails.title.EN",
        }
    }];
    query.joinWithAnd(dbConstants.dbSchema.properties, joinArr, async (error, response) => {
        if (error) {
            logger('Error: can not get record.');
            done(errors.internalServer(true), null);
            return;
        }
        _.each(response, (elem) => {
            let arr = []
            _.each(elem.parking_spot_photos, (rec) => {
                if (rec != '') {
                    arr.push(config.aws.prefix + config.aws.s3.propertyBucket + '/' + rec)
                }
            });
            elem.parking_spot_photos = arr
        })
        done(null, response)
    });
}

const getEarning = async (requestParam, done) => {
    let promise = [];
    let matchColumn = {
        owner_id: requestParam.customer_id,
        status: 'completed'
    }
    if (requestParam.property_id) {
        matchColumn.property_id = requestParam.property_id
    }
    if (requestParam.type == 'daily') {
        var d = new Date();
        var dayName = days[d.getDay()]
        let date = moment(d).format('YYYY-MM-DD')
        matchColumn.completed_date = {
            $lte: new Date(date + 'T23:59:59.000Z'),
            $gte: new Date(date + 'T00:00:00.000Z')
        }
        let data = await queryApi.selectWithAnd(dbConstants.dbSchema.reservations, matchColumn, { _id: 0, owner_earn: 1, });
        let owner_earn = parseFloat(LD.sumBy(data, 'owner_earn')).toFixed(2)
        owner_earn = parseFloat(owner_earn)
        let obj = { label: [dayName], value: [owner_earn], total_earn: owner_earn }
        done(null, obj)
        return;
    }
    else if (requestParam.type == 'weekly') {
        let start_date = moment().startOf('week').format('YYYY-MM-DD')
        let end_date = moment().endOf('week').format('YYYY-MM-DD')

        let weekName = []
        var getDaysArray = function (start, end) {
            for (var arr = [], dt = new Date(start); dt <= end; dt.setDate(dt.getDate() + 1)) {
                arr.push(moment(new Date(dt)).format('YYYY-MM-DD'));
                weekName.push(days[dt.getDay()])
            }
            return arr;
        };
        var dateArr = getDaysArray(new Date(start_date), new Date(end_date));
        for (let x in weekName) {
            promise.push(await weeklyEarning(weekName[x], dateArr[x], matchColumn))
        }
        Promise.all(promise)
            .then(async result => {
                done(null, {
                    label: _.pluck(result, 'label'),
                    value: _.pluck(result, 'value'),
                    total_earn: parseFloat(LD.sumBy(result, 'value')).toFixed(2)
                })
            });
    }
    else if (requestParam.type == 'monthly') {
        let start_date = moment().startOf('month').format('YYYY-MM-DD')
        let end_date = moment().endOf('month').format('YYYY-MM-DD')
        for (let x in monthName) {
            promise.push(await monthlyEarning(monthName[x], x, matchColumn))
        }
        Promise.all(promise)
            .then(async result => {
                done(null, {
                    label: _.pluck(result, 'label'),
                    value: _.pluck(result, 'value'),
                    total_earn: parseFloat(LD.sumBy(result, 'value')).toFixed(2)
                })
            });
    }
    else if (requestParam.type == 'quarterly') {
        let start_date = moment().startOf('year').format('YYYY-MM-DD')
        let end_date = moment().endOf('year').format('YYYY-MM-DD')
        let q = [0, 3, 6, 9];
        for (let x in q) {
            promise.push(await quarterlyEarning(q[x], matchColumn))
        }
        Promise.all(promise)
            .then(async result => {
                done(null, {
                    label: _.pluck(result, 'label'),
                    value: _.pluck(result, 'value'),
                    total_earn: parseFloat(LD.sumBy(result, 'value')).toFixed(2)
                })
            });
    }
};

const weeklyEarning = async (lbl, date, matchColumn) => {
    return new Promise(async (resolve, reject) => {
        try {
            matchColumn.completed_date = {
                $lte: new Date(date + 'T23:59:59.000Z'),
                $gte: new Date(date + 'T00:00:00.000Z')
            }
            let data = await queryApi.selectWithAnd(dbConstants.dbSchema.reservations, matchColumn, { _id: 0, price: 1, });
            let price = parseFloat(LD.sumBy(data, 'price')).toFixed(2)
            price = parseFloat(price)
            let obj = { label: lbl, value: price }
            resolve(obj)
            return;
        } catch (error) {
            console.log(error)
            reject(error)
            return
        }
    })
};

const monthlyEarning = async (lbl, index, matchColumn) => {
    return new Promise(async (resolve, reject) => {
        try {
            var date = new Date();
            let y = date.getFullYear();
            let m = parseInt(index);
            var firstDay = new Date(y, m, 1);
            firstDay.setHours(0, 0, 0, 0);
            var lastDay = new Date(y, m + 1, 0);
            lastDay.setHours(23, 59, 59, 999);
            matchColumn.completed_date = {
                $lte: lastDay,
                $gte: firstDay
            }
            let data = await queryApi.selectWithAnd(dbConstants.dbSchema.reservations, matchColumn, { _id: 0, price: 1, });
            let price = parseFloat(LD.sumBy(data, 'price')).toFixed(2)
            price = parseFloat(price)
            let obj = { label: lbl, value: price }
            resolve(obj)
            return;
        } catch (error) {
            console.log(error)
            reject(error)
            return
        }
    })
};

const quarterlyEarning = async (index, matchColumn) => {
    return new Promise(async (resolve, reject) => {
        try {
            var date = new Date();
            let y = date.getFullYear();
            let m = parseInt(index);
            var firstDay = new Date(y, m, 1);
            firstDay.setHours(0, 0, 0, 0);
            m = m + 2
            var lastDay = new Date(y, m + 1, 0);
            lastDay.setHours(23, 59, 59, 999);
            matchColumn.completed_date = {
                $lte: lastDay,
                $gte: firstDay
            }
            let data = await queryApi.selectWithAnd(dbConstants.dbSchema.reservations, matchColumn, { _id: 0, price: 1, });
            let price = parseFloat(LD.sumBy(data, 'price')).toFixed(2)
            price = parseFloat(price)
            let lbl = 'Jan-Mar';
            if (index == 3) {
                lbl = 'Apr-Jun';
            }
            if (index == 6) {
                lbl = 'Jul-Sep';
            }
            if (index == 9) {
                lbl = 'Oct-Dec';
            }
            let obj = { label: lbl, value: price }
            resolve(obj)
            return;
        } catch (error) {
            console.log(error)
            reject(error)
            return
        }
    })
};

const ratings = async (requestParam, done) => {
    let joinArr = [{
        $lookup: {
            from: 'customers',
            localField: 'customer_id',
            foreignField: 'customer_id',
            as: 'cusDetails'
        }
    }, {
        $unwind: "$cusDetails"
    }, {
        $match: {
            customer_id: requestParam.customer_id,
        }
    }, {
        $sort: {
            created_at: -1
        }
    }, {
        $project: {
            _id: 0,
            name: "$cusDetails.name",
            rating: "$rating",
            rating_status: "$rating_status",
            created_at: "$created_at",
        }
    }];
    let gave_to_owner_ratings = await queryApi.joinWithAnd(dbConstants.dbSchema.owner_ratings, joinArr);
    _.each(gave_to_owner_ratings, (elem) => {
        elem.created_at = timeZone(new Date(elem.created_at)).tz(requestParam.time_zone).format('lll');
    })

    joinArr = [{
        $lookup: {
            from: 'customers',
            localField: 'customer_id',
            foreignField: 'customer_id',
            as: 'cusDetails'
        }
    }, {
        $unwind: "$cusDetails"
    }, {
        $match: {
            owner_id: requestParam.customer_id,
        }
    }, {
        $sort: {
            created_at: -1
        }
    }, {
        $project: {
            _id: 0,
            name: "$cusDetails.name",
            rating: "$rating",
            rating_status: "$rating_status",
            created_at: "$created_at",
        }
    }];

    let received_from_seeker_ratings = await queryApi.joinWithAnd(dbConstants.dbSchema.owner_ratings, joinArr);
    _.each(received_from_seeker_ratings, (elem) => {
        elem.created_at = timeZone(new Date(elem.created_at)).tz(requestParam.time_zone).format('lll');
    })
    done(null, { gave_to_owner_ratings, received_from_seeker_ratings })
};

module.exports = {
    get,
    create,
    action,
    update,
    getBankDetails,
    getPropertyList,
    getEarning,
    ratings
};