'use strict';

const logger = require('./../../utils/logger');
const jsonResponse = require('./../../utils/json-response');
const errors = require('./../../utils/dz-errors');
const config = require('./../../config');
const dbConstants = require('./../../constants/db-constants');
const query = require('./../../utils/query-creator');
let async = require('async');
let _ = require('underscore');
let moment = require('moment');
const timeZone = require('moment-timezone');

const get = function(requestParam,done){
    let columnAndValue = {};
    if(requestParam.keyword && requestParam.keyword!=''){
        columnAndValue.address = new RegExp(requestParam.keyword, 'i')
    }
    if(requestParam.start_date && requestParam.end_date){
        columnAndValue.from_date_range = {
            $lte: new Date(requestParam.end_date+'T23:59:59.000Z'),
        }
        columnAndValue.to_date_range = {
            $gte: new Date(requestParam.start_date+'T00:00:00.000Z')
        }
    }
    let joinArr = [{
        $lookup: {
            from: 'parking_spot_coverages',
            localField: 'parking_spot_coverage_id',
            foreignField: 'parking_spot_coverage_id',
            as: 'spotCoverageDetails'
        }
    },  {
        $unwind: "$spotCoverageDetails"
    },  {
        $lookup: {
            from: 'parking_spot_types',
            localField: 'parking_spot_type_id',
            foreignField: 'parking_spot_type_id',
            as: 'spotTypesDetails'
        }
    },  {
        $unwind: "$spotTypesDetails"
    },  {
        $match: columnAndValue
    }, { 
        $sort : {created_at:-1}
    }, {
        $project: {
            _id: 0,
            property_id:1,
            address:1,
            parking_spot_photos:1,
            pricing:1,
            status:1,
            is_specific_date_range:1,
            from_date_range:1,
            to_date_range:1,
            parking_spot_type_title: "$spotTypesDetails.title.EN",
            parking_spot_coverage_title: "$spotCoverageDetails.title.EN",
        }
    }];
    query.joinWithAnd(dbConstants.dbSchema.properties, joinArr, async (error, response) => {
        if (error) {
            logger('Error: can not get record.');
            done(errors.internalServer(true), null);
            return;
        }
        response = JSON.parse(JSON.stringify(response))
        _.each(response, (elem) => {
            let arr = []
            _.each(elem.parking_spot_photos, (rec) => {
                if(rec != ''){
                    arr.push(config.aws.prefix + config.aws.s3.propertyBucket + '/' + rec)
                }
            });
            elem.parking_spot_photos = arr
            if(!elem.pricing || !elem.pricing.hourly_rate){
                elem.pricing = {
                    hourly_rate:0
                }
            }
            if(elem.is_specific_date_range == true){
                elem.is_available = true;
            }
            else{
                let todayDate = timeZone(new Date()).tz(requestParam.time_zone);
                if(moment(todayDate).isBetween(elem.from_date_range, elem.to_date_range, null, '[]')){
                    elem.is_available = true;
                }
                else{
                    elem.is_available = false;
                }
            }
        })
        done(null, response)
    });
};

const action  = (requestParam, done) => {
    if (requestParam['type']=="delete") {
        query.removeMultiple(dbConstants.dbSchema.properties, {
            'property_id': {
                $in: requestParam['ids']
            }
        }, function(error, data) {
            if (error) {
                logger('Error: can not delete ');
                done(error, null);
                return;
            }
            done(null, data);
        });        
    }
    else
    {
        done(null, {});
    }
};

const create  = (requestParam, done) => {
    query.insertSingle(dbConstants.dbSchema.properties,requestParam,function (error, property) {
        if (error) {
            logger('Error: can not create property');
            done(error, null);
            return;
        }
        console.log(property.property_id)
        done(null, {});
    });
};

module.exports = {
	get,
    action,
    create
};